# SPDX-License-Identifier: Apache-2.0

from logging import getLogger
from collections import OrderedDict
import numpy as np
from ..common.data_types import FloatTensorType, DoubleTensorType
from ..common.utils import get_unique_subgraph
from .onnx_ops import (
    OnnxAbs,
    OnnxDiv,
    OnnxIdentity,
    OnnxMatMul,
    OnnxPow,
    OnnxScan,
    OnnxSqrt,
    OnnxSub,
    OnnxReduceSumApi11,
    OnnxReduceSumSquareApi18,
    OnnxTranspose,
)

logger = getLogger("skl2onnx")


def onnx_squareform_pdist(
    X, metric="sqeuclidean", dtype=None, op_version=None, **kwargs
):
    """
    Returns the ONNX graph which computes
    ``squareform(pdist(X, metric=metric))``.
    """
    if metric == "sqeuclidean":
        return _onnx_squareform_pdist_sqeuclidean(
            X, dtype=dtype, op_version=op_version, **kwargs
        )
    if metric == "euclidean":
        res = _onnx_squareform_pdist_sqeuclidean(X, dtype=dtype, op_version=op_version)
        return OnnxSqrt(res, op_version=op_version, **kwargs)
    raise NotImplementedError("metric='{}' is not implemented.".format(metric))


def _onnx_squareform_pdist_sqeuclidean(X, dtype=None, op_version=None, **kwargs):
    """
    Returns the ONNX graph which computes
    ``squareform(pdist(X, metric='sqeuclidean'))``.
    """
    unique = get_unique_subgraph()
    diff = OnnxSub("next_in", "next", op_version=op_version)
    id_next = OnnxIdentity("next_in", output_names=["next_out"], op_version=op_version)
    flat = OnnxReduceSumSquareApi18(
        diff, axes=[1], op_version=op_version, output_names=["scan_out"], keepdims=0
    )
    flat.set_onnx_name_prefix("cflat_%d" % unique)
    id_next.set_onnx_name_prefix("pdistsqe_%d" % unique)
    tensor_type = FloatTensorType if dtype == np.float32 else DoubleTensorType
    scan_body = id_next.to_onnx(
        OrderedDict(
            [("next_in", tensor_type([None, None])), ("next", tensor_type([None]))]
        ),
        outputs=[
            ("next_out", tensor_type([None, None])),
            ("scan_out", tensor_type([None])),
        ],
        other_outputs=[flat],
        target_opset=op_version,
    )

    node = OnnxScan(
        X,
        X,
        output_names=["u(scan0)", "u(scan1)"],
        num_scan_inputs=1,
        body=(scan_body.graph, [id_next, flat]),
        op_version=op_version,
        **kwargs,
    )
    logger.debug("[_onnx_squareform_pdist_sqeuclidean] +Scan dtype=%r", dtype)
    return node[1]


def onnx_cdist(
    XA,
    XB,
    metric="sqeuclidean",
    dtype=None,
    op_version=None,
    dim_in=None,
    dim_out=None,
    **kwargs,
):
    """
    Returns the ONNX graph which computes
    ``cdist(XA, XB, metric=metric)``.

    :param XA: array or OnnxOperatorMixin
    :param XB: array or OnnxOperatorMixin
    :param metric: distance type
    :param dtype: *np.float32* or *np.float64*
    :param op_version: opset version
    :param dim_in: dimension of the input vectorial space
        (if known)
    :param dim_out: dimension of the output vectorial space
        (if known)
    :param kwargs: addition parameter
    :return: OnnxOperatorMixin
    """
    if metric == "sqeuclidean":
        return _onnx_cdist_sqeuclidean(
            XA,
            XB,
            dtype=dtype,
            op_version=op_version,
            dim_in=dim_in,
            dim_out=dim_out,
            **kwargs,
        )
    if metric == "euclidean":
        res = _onnx_cdist_sqeuclidean(
            XA, XB, dtype=dtype, op_version=op_version, dim_in=dim_in, dim_out=dim_out
        )
        return OnnxSqrt(res, op_version=op_version, **kwargs)
    if metric == "minkowski":
        p = kwargs.pop("p")
        res = _onnx_cdist_minkowski(
            XA,
            XB,
            dtype=dtype,
            op_version=op_version,
            p=p,
            dim_in=dim_in,
            dim_out=dim_out,
        )
        return OnnxPow(
            res, np.array([1.0 / p], dtype=dtype), op_version=op_version, **kwargs
        )
    if metric in ("manhattan", "cityblock"):
        return _onnx_cdist_manhattan(
            XA,
            XB,
            dtype=dtype,
            op_version=op_version,
            dim_in=dim_in,
            dim_out=dim_out,
            **kwargs,
        )
    if metric == "cosine":
        return _onnx_cdist_cosine(
            XA,
            XB,
            dtype=dtype,
            op_version=op_version,
            dim_in=dim_in,
            dim_out=dim_out,
            **kwargs,
        )
    raise NotImplementedError(f"metric={metric!r} is not implemented.")


def _onnx_cdist_begin(op_version):
    diff = OnnxSub("next_in", "next", op_version=op_version)
    id_next = OnnxIdentity("next_in", output_names=["next_out"], op_version=op_version)
    return diff, id_next


def _onnx_cdist_end(
    XA, XB, id_next, flat, dtype, op_version, dim_in=None, dim_out=None, **kwargs
):
    unique = get_unique_subgraph()
    tensor_type = FloatTensorType if dtype == np.float32 else DoubleTensorType
    id_next.set_onnx_name_prefix("cdistd_%d" % unique)
    flat.set_onnx_name_prefix("cdistdf_%d" % unique)
    shape_in = (
        tensor_type([None, None]) if dim_in is None else tensor_type([None, dim_in])
    )
    scan_body = id_next.to_onnx(
        OrderedDict([("next_in", shape_in), ("next", tensor_type([None]))]),
        outputs=[
            ("next_out", tensor_type([None, None])),
            ("scan_out", tensor_type([None])),
        ],
        other_outputs=[flat],
        target_opset=op_version,
    )
    logger.debug(
        "[_onnx_cdist_end] + Scan dim_in=%r dim_out=%r dtype=%r", dim_in, dim_out, dtype
    )

    node = OnnxScan(
        XA,
        XB,
        output_names=["u(scan0)", "u(scan1)"],
        num_scan_inputs=1,
        body=(scan_body.graph, [id_next, flat]),
        op_version=op_version,
    )
    return OnnxTranspose(node[1], perm=[1, 0], op_version=op_version, **kwargs)


def _onnx_cdist_sqeuclidean(
    XA, XB, dtype=None, op_version=None, dim_in=None, dim_out=None, **kwargs
):
    """
    Returns the ONNX graph which computes
    ``cdist(X, metric='sqeuclidean')``.
    """
    diff, id_next = _onnx_cdist_begin(op_version)
    norm = OnnxReduceSumSquareApi18(diff, axes=[1], keepdims=0, op_version=op_version)
    flat = OnnxIdentity(norm, output_names=["scan_out"], op_version=op_version)
    return _onnx_cdist_end(
        XA,
        XB,
        id_next,
        flat,
        dtype,
        op_version,
        dim_in=dim_in,
        dim_out=dim_out,
        **kwargs,
    )


def _onnx_cdist_minkowski(
    XA, XB, dtype=None, op_version=None, p=2, dim_in=None, dim_out=None, **kwargs
):
    """
    Returns the ONNX graph which computes the Minkowski distance
    or ``minkowski(XA, XB, p)``.
    """
    diff, id_next = _onnx_cdist_begin(op_version)
    diff_pow = OnnxPow(
        OnnxAbs(diff, op_version=op_version),
        np.array([p], dtype=dtype),
        op_version=op_version,
    )
    norm = OnnxReduceSumApi11(diff_pow, axes=[1], keepdims=0, op_version=op_version)
    norm.set_onnx_name_prefix("norm_%d" % id(norm))
    flat = OnnxIdentity(norm, output_names=["scan_out"], op_version=op_version)
    return _onnx_cdist_end(
        XA,
        XB,
        id_next,
        flat,
        dtype,
        op_version,
        dim_in=dim_in,
        dim_out=dim_out,
        **kwargs,
    )


def _onnx_cdist_manhattan(
    XA, XB, dtype=None, op_version=None, dim_in=None, dim_out=None, **kwargs
):
    """
    Returns the ONNX graph which computes the Manhattan distance
    or ``Manhattan(X, Y)``.
    """
    diff, id_next = _onnx_cdist_begin(op_version)
    diff_pow = OnnxAbs(diff, op_version=op_version)
    norm = OnnxReduceSumApi11(diff_pow, axes=[1], keepdims=0, op_version=op_version)
    norm.set_onnx_name_prefix("norm_%d" % id(norm))
    flat = OnnxIdentity(norm, output_names=["scan_out"], op_version=op_version)
    return _onnx_cdist_end(
        XA,
        XB,
        id_next,
        flat,
        dtype,
        op_version,
        dim_in=dim_in,
        dim_out=dim_out,
        **kwargs,
    )


def _onnx_cdist_cosine(
    XA, XB, dtype=None, op_version=None, dim_in=None, dim_out=None, **kwargs
):
    """
    Returns the ONNX graph which computes
    ``cdist(X, metric='cosine')``.
    """
    txb = OnnxTranspose(XB, perm=[1, 0], op_version=op_version)
    scal = OnnxMatMul(XA, txb, op_version=op_version)
    norma = OnnxSqrt(
        OnnxReduceSumSquareApi18(XA, axes=[1], keepdims=1, op_version=op_version),
        op_version=op_version,
    )
    normb = OnnxSqrt(
        OnnxReduceSumSquareApi18(txb, axes=[0], keepdims=1, op_version=op_version),
        op_version=op_version,
    )
    return OnnxSub(
        np.array([1], dtype=dtype),
        OnnxDiv(
            scal,
            OnnxMatMul(norma, normb, op_version=op_version),
            op_version=op_version,
        ),
        op_version=op_version,
        **kwargs,
    )



# SPDX-License-Identifier: Apache-2.0


from ..common._apply_operation import (
    apply_add,
    apply_argmax,
    apply_cast,
    apply_concat,
    apply_div,
    apply_exp,
    apply_log,
    apply_matmul,
    apply_mul,
    apply_pow,
    apply_reducesum,
    apply_reshape,
    apply_sub,
    apply_transpose,
)
from ..common.data_types import BooleanTensorType, Int64TensorType, guess_proto_type
from ..common._registration import register_converter
from ..common._topology import Scope, Operator
from ..common._container import ModelComponentContainer
from ..proto import onnx_proto


def convert_quadratic_discriminant_analysis_classifier(
    scope: Scope, operator: Operator, container: ModelComponentContainer
):
    input_name = operator.inputs[0].full_name
    model = operator.raw_operator

    n_classes = len(model.classes_)

    proto_dtype = guess_proto_type(operator.inputs[0].type)
    if proto_dtype != onnx_proto.TensorProto.DOUBLE:
        proto_dtype = onnx_proto.TensorProto.FLOAT

    if isinstance(operator.inputs[0].type, (BooleanTensorType, Int64TensorType)):
        cast_input_name = scope.get_unique_variable_name("cast_input")
        apply_cast(
            scope, operator.input_full_names, cast_input_name, container, to=proto_dtype
        )
        input_name = cast_input_name

    norm_array_name = []
    sum_array_name = []

    const_n05 = scope.get_unique_variable_name("const_n05")
    const_p2 = scope.get_unique_variable_name("const_p2")
    container.add_initializer(const_n05, proto_dtype, [], [-0.5])
    container.add_initializer(const_p2, proto_dtype, [], [2])

    for i in range(n_classes):
        R = model.rotations_[i]
        rotation_name = scope.get_unique_variable_name("rotations")
        container.add_initializer(
            rotation_name, proto_dtype, [R.shape[0], R.shape[1]], R
        )

        S = model.scalings_[i]
        scaling_name = scope.get_unique_variable_name("scalings")
        container.add_initializer(
            scaling_name,
            proto_dtype,
            [
                S.shape[0],
            ],
            S,
        )

        mean = model.means_[i]
        mean_name = scope.get_unique_variable_name("means")
        container.add_initializer(mean_name, proto_dtype, mean.shape, mean)

        Xm_name = scope.get_unique_variable_name("Xm")
        apply_sub(scope, [input_name, mean_name], [Xm_name], container)

        s_pow_name = scope.get_unique_variable_name("s_pow_n05")
        apply_pow(scope, [scaling_name, const_n05], [s_pow_name], container)

        mul_name = scope.get_unique_variable_name("mul")
        apply_mul(scope, [rotation_name, s_pow_name], [mul_name], container)

        x2_name = scope.get_unique_variable_name("matmul")
        apply_matmul(scope, [Xm_name, mul_name], [x2_name], container)

        pow_x2_name = scope.get_unique_variable_name("pow_x2")
        apply_pow(scope, [x2_name, const_p2], [pow_x2_name], container)

        sum_name = scope.get_unique_variable_name("sum")
        apply_reducesum(
            scope, [pow_x2_name], [sum_name], container, axes=[1], keepdims=1
        )
        norm_array_name.append(sum_name)

        log_name = scope.get_unique_variable_name("log")
        apply_log(scope, [scaling_name], [log_name], container)

        sum_log_name = scope.get_unique_variable_name("sum_log")
        apply_reducesum(scope, [log_name], [sum_log_name], container, keepdims=1)
        sum_array_name.append(sum_log_name)

    concat_norm_name = scope.get_unique_variable_name("concat_norm")
    apply_concat(scope, norm_array_name, [concat_norm_name], container)

    reshape_norm_name = scope.get_unique_variable_name("reshape_concat_norm")
    apply_reshape(
        scope,
        [concat_norm_name],
        [reshape_norm_name],
        container,
        desired_shape=[n_classes, -1],
    )

    transpose_norm_name = scope.get_unique_variable_name("transpose_norm")
    apply_transpose(
        scope, [reshape_norm_name], [transpose_norm_name], container, perm=(1, 0)
    )

    concat_logsum = scope.get_unique_variable_name("concat_logsum")
    apply_concat(scope, sum_array_name, [concat_logsum], container)

    add_norm2_u_name = scope.get_unique_variable_name("add_norm2_u")
    apply_add(
        scope, [transpose_norm_name, concat_logsum], [add_norm2_u_name], container
    )

    norm2_u_n05_name = scope.get_unique_variable_name("norm2_u_n05")
    apply_mul(scope, [const_n05, add_norm2_u_name], [norm2_u_n05_name], container)

    priors = scope.get_unique_variable_name("priors")
    container.add_initializer(
        priors,
        proto_dtype,
        [
            n_classes,
        ],
        model.priors_,
    )
    log_p = scope.get_unique_variable_name("log_p")
    apply_log(scope, [priors], [log_p], container)

    decision_fun = scope.get_unique_variable_name("decision_fun")
    apply_add(scope, [norm2_u_n05_name, log_p], [decision_fun], container)

    argmax_out = scope.get_unique_variable_name("argmax_out")
    apply_argmax(scope, [decision_fun], [argmax_out], container, axis=1)

    classes = scope.get_unique_variable_name("classes")
    container.add_initializer(
        classes, onnx_proto.TensorProto.INT64, [n_classes], model.classes_
    )

    container.add_node(
        "ArrayFeatureExtractor",
        [classes, argmax_out],
        [operator.outputs[0].full_name],
        op_domain="ai.onnx.ml",
    )

    df_max = scope.get_unique_variable_name("df_max")
    if container.target_opset >= 18:
        axis_name = scope.get_unique_variable_name("axis")
        container.add_initializer(axis_name, onnx_proto.TensorProto.INT64, [1], [1])
        container.add_node("ReduceMax", [decision_fun, axis_name], [df_max])
    else:
        container.add_node("ReduceMax", [decision_fun], [df_max], axes=[1])
    df_sub_max = scope.get_unique_variable_name("df_sub_max")
    apply_sub(scope, [decision_fun, df_max], [df_sub_max], container)
    likelihood = scope.get_unique_variable_name("likelihood")
    apply_exp(scope, [df_sub_max], [likelihood], container)
    likelihood_sum = scope.get_unique_variable_name("likelihood_sum")
    apply_reducesum(
        scope, [likelihood], [likelihood_sum], container, axes=[1], keepdims=1
    )
    apply_div(
        scope,
        [likelihood, likelihood_sum],
        [operator.outputs[1].full_name],
        container,
    )


register_converter(
    "SklearnQuadraticDiscriminantAnalysis",
    convert_quadratic_discriminant_analysis_classifier,
    options={
        "zipmap": [True, False, "columns"],
        "nocl": [True, False],
        "output_class_labels": [False, True],
    },
)



# SPDX-License-Identifier: Apache-2.0


from .investigate import collect_intermediate_steps, compare_objects  # noqa
from .investigate import enumerate_pipeline_models  # noqa
from .integration import add_onnx_graph  # noqa



# SPDX-License-Identifier: Apache-2.0

from ..common._registration import register_shape_calculator
from ..common.data_types import Int64TensorType


def calculate_quadratic_discriminant_analysis_shapes(operator):
    N = len(operator.raw_operator.classes_)
    operator.outputs[0].type = Int64TensorType([1, N])
    operator.outputs[1].type.shape = [None, N]


register_shape_calculator(
    "SklearnQuadraticDiscriminantAnalysis",
    calculate_quadratic_discriminant_analysis_shapes,
)



# SPDX-License-Identifier: Apache-2.0

import warnings
from sklearn.base import BaseEstimator
from onnx import shape_inference
from ..common._topology import Scope, Operator
from ..common._container import ModelComponentContainer
from ..common._registration import get_converter, get_shape_calculator
from ..common._topology import Variable
from .._supported_operators import sklearn_operator_name_map
from .onnx_operator import OnnxOperator
from .type_helper import guess_initial_types


class OnnxOperatorMixin:
    """
    Base class for *scikit-learn* operators
    sharing an API to convert object to *ONNX*.
    """

    def to_onnx(
        self,
        X=None,
        name=None,
        options=None,
        white_op=None,
        black_op=None,
        final_types=None,
        target_opset=None,
        verbose=0,
    ):
        """
        Converts the model in *ONNX* format.
        It calls method *_to_onnx* which must be
        overloaded.

        :param X: training data, at least one sample,
            it is used to guess the type of the input data.
        :param name: name of the model, if None,
            it is replaced by the the class name.
        :param options: specific options given to converters
            (see :ref:`l-conv-options`)
        :param white_op: white list of ONNX nodes allowed
            while converting a pipeline, if empty, all are allowed
        :param black_op: black list of ONNX nodes allowed
            while converting a pipeline, if empty, none are blacklisted
        :param final_types: a python list. Works the same way as initial_types
            but not mandatory, it is used to overwrites the type
            (if type is not None) and the name of every output.
        :param target_opset: to overwrite `self.op_version`
        :param verbose: displays information while converting
        """
        from .. import convert_sklearn

        if X is None:
            initial_types = self.infer_initial_types()
        else:
            initial_types = guess_initial_types(X, None)
        if not hasattr(self, "op_version"):
            if name is None:
                name = self.__class__.__name__
            raise AttributeError(
                "Attribute 'op_version' is missing for '{}' "
                "(model: '{}').".format(self.__class__.__name__, name)
            )
        return convert_sklearn(
            self,
            initial_types=initial_types,
            target_opset=target_opset or self.op_version,
            options=options,
            white_op=white_op,
            black_op=black_op,
            final_types=final_types,
            verbose=verbose,
        )

    def infer_initial_types(self):
        """
        Infers initial types.
        """
        if hasattr(self, "enumerate_initial_types"):
            return list(self.enumerate_initial_types())
        raise RuntimeError(
            "Method enumerate_initial_types is missing "
            "and initial_types are not defined."
        )

    def _find_sklearn_parent(self):
        for cl in self.__class__.__bases__:
            if issubclass(cl, BaseEstimator):
                return cl
        raise RuntimeError(
            "Unable to find any parent inherited from "
            "BaseEstimator: {}.".format(", ".join(map(str, self.__class__.__bases__)))
        )

    def to_onnx_operator(
        self, inputs=None, outputs=None, target_opset=None, options=None
    ):
        """
        This function must be overloaded.
        """
        raise NotImplementedError()

    def onnx_parser(self):
        """
        Returns a parser for this model.
        If not overloaded, it calls the converter to guess the number
        of outputs. If it still fails, it fetches the parser
        mapped to the first *scikit-learn* parent
        it can find.
        """

        def parser(scope=None, inputs=None):
            try:
                op = self.to_onnx_operator(inputs=inputs, outputs=None)
            except NotImplementedError:
                self._find_sklearn_parent()
                return None

            names = []
            while True:
                try:
                    name = op.get_output_name(len(names), scope=scope)
                    if name is None:
                        break
                    names.append(name)
                except AttributeError:
                    return None
                except IndexError:
                    break
            return names

        return parser

    def get_inputs(self, inputs, i):
        if i >= len(inputs):
            return OnnxOperator.OnnxOperatorVariable(i)
        else:
            input = inputs[i]
            if isinstance(input, (str, OnnxOperator.UnscopedVariable)):
                return OnnxOperator.OnnxOperatorVariable(i, input)
            else:
                return input

    def onnx_shape_calculator(self):
        """
        Returns a shape calculator for this model.
        If not overloaded, it fetches the parser
        mapped to the first *scikit-learn* parent
        it can find.
        """
        if not hasattr(self, "op_version"):
            raise AttributeError(
                "Class '{}' should have an attribute 'op_version'.".format(
                    self.__class__.__name__
                )
            )

        try:
            op = self.to_onnx_operator()
        except NotImplementedError:
            parent = self._find_sklearn_parent()
            name = sklearn_operator_name_map.get(parent, "Sklearn" + parent.__name__)
            return get_shape_calculator(name)

        def shape_calculator(operator):
            onx = op.to_onnx(
                operator.inputs, operator.outputs, target_opset=self.op_version
            )
            inferred_model = shape_inference.infer_shapes(onx)
            shapes = Variable.from_pb(inferred_model.graph.value_info)
            shapes = {shape.onnx_name: shape for shape in shapes}
            for o in operator.outputs:
                name = o.onnx_name
                if name not in shapes:
                    raise RuntimeError(
                        "Shape of output '{}' cannot be "
                        "infered. onnx_shape_calculator "
                        "must be overriden and return "
                        "a shape calculator.".format(name)
                    )
                o.set_type(shapes[name].type)

        return shape_calculator

    def onnx_converter(self):
        """
        Returns a converter for this model.
        If not overloaded, it fetches the converter
        mapped to the first *scikit-learn* parent
        it can find.
        """

        def converter(
            scope: Scope, operator: Operator, container: ModelComponentContainer
        ):
            inputs = operator.inputs  # getattr(self, "parsed_inputs_", None)
            outputs = operator.outputs  # kwargs.get('outputs', None)
            op_version = container.target_opset
            options = scope.get_options(operator.raw_operator, fail=False)
            try:
                if inputs:
                    op = self.to_onnx_operator(
                        inputs=inputs,
                        outputs=outputs,
                        target_opset=op_version,
                        options=options,
                    )
                else:
                    op = self.to_onnx_operator(
                        target_opset=op_version, outputs=outputs, options=options
                    )
            except TypeError:
                warnings.warn(
                    "Signature should be to_onnx_operator(self, inputs=None, "
                    "outputs=None, target_opset=None, **kwargs). "
                    "This will be the case in version 1.11, class=%r."
                    "" % type(self),
                    DeprecationWarning,
                )
                try:
                    if inputs:
                        op = self.to_onnx_operator(inputs=inputs, outputs=outputs)
                    else:
                        op = self.to_onnx_operator()
                except NotImplementedError:
                    parent = self._find_sklearn_parent()
                    name = sklearn_operator_name_map[parent]
                    conv = get_converter(name)
                    return conv(scope, operator, container)
            except NotImplementedError:
                parent = self._find_sklearn_parent()
                name = sklearn_operator_name_map[parent]
                conv = get_converter(name)
                return conv(scope, operator, container)

            op.add_to(scope, container, operator=operator)

        return converter



# SPDX-License-Identifier: Apache-2.0


import numpy as np
from ._apply_operation import apply_cast, apply_reshape
from ..proto import onnx_proto


def get_label_classes(scope, op, node_names=False):
    """
    Extracts the model classes,
    handles option ``nocl`` and ``zipmap=='columns'``
    """
    options = scope.get_options(op, dict(nocl=False))
    if options["nocl"]:
        if len(op.classes_.shape) > 1 and op.classes_.shape[1] > 1:
            raise RuntimeError(
                "Options 'nocl=True' is not implemented for multi-label "
                "classification (class: {}).".format(op.__class__.__name__)
            )
        classes = np.arange(0, len(op.classes_))
    elif node_names:
        try:
            options = scope.get_options(op, dict(zipmap=False))
            zipcol = options["zipmap"] == "columns"
        except NameError:
            zipcol = False
        if zipcol:
            clnames = op.classes_.ravel()
            if np.issubdtype(clnames.dtype, np.integer) or clnames.dtype == np.bool_:
                classes = np.array(["i%d" % c for c in clnames])
            else:
                classes = np.array(["s%s" % c for c in clnames])
        else:
            classes = op.classes_
    elif hasattr(op, "classes_"):
        classes = op.classes_
    elif hasattr(op, "intercept_"):
        classes = len(op.intercept_)
    elif hasattr(op, "y_"):
        # _ConstantPredictor
        classes = np.array(list(sorted(set(op.y_))))
    else:
        raise RuntimeError(
            "No known ways to retrieve the number of classes for class %r."
            "" % type(op)
        )
    return classes


def _finalize_converter_classes(
    scope, argmax_output_name, output_full_name, container, classes, proto_dtype
):
    """
    See :func:`convert_voting_classifier`.
    """
    if np.issubdtype(classes.dtype, np.floating) or classes.dtype == np.bool_:
        class_type = onnx_proto.TensorProto.INT32
        classes = np.array(list(map(lambda x: int(x), classes)))
    elif np.issubdtype(classes.dtype, np.signedinteger):
        class_type = onnx_proto.TensorProto.INT32
    else:
        classes = np.array([s.encode("utf-8") for s in classes])
        class_type = onnx_proto.TensorProto.STRING

    classes_name = scope.get_unique_variable_name("classes")
    container.add_initializer(classes_name, class_type, classes.shape, classes)

    array_feature_extractor_result_name = scope.get_unique_variable_name(
        "array_feature_extractor_result"
    )
    container.add_node(
        "ArrayFeatureExtractor",
        [classes_name, argmax_output_name],
        array_feature_extractor_result_name,
        op_domain="ai.onnx.ml",
        name=scope.get_unique_operator_name("ArrayFeatureExtractor"),
    )

    output_shape = (-1,)
    if class_type == onnx_proto.TensorProto.INT32:
        cast2_result_name = scope.get_unique_variable_name("cast2_result")
        reshaped_result_name = scope.get_unique_variable_name("reshaped_result")
        apply_cast(
            scope,
            array_feature_extractor_result_name,
            cast2_result_name,
            container,
            to=proto_dtype,
        )
        apply_reshape(
            scope,
            cast2_result_name,
            reshaped_result_name,
            container,
            desired_shape=output_shape,
        )
        apply_cast(
            scope,
            reshaped_result_name,
            output_full_name,
            container,
            to=onnx_proto.TensorProto.INT64,
        )
    else:  # string labels
        apply_reshape(
            scope,
            array_feature_extractor_result_name,
            output_full_name,
            container,
            desired_shape=output_shape,
        )



# SPDX-License-Identifier: Apache-2.0
import copy

from ..common._registration import register_shape_calculator


def calculate_sklearn_multiply(operator):
    for variable, output in zip(operator.inputs, operator.outputs):
        output.type = copy.copy(variable.type)


register_shape_calculator("SklearnMultiply", calculate_sklearn_multiply)



# SPDX-License-Identifier: Apache-2.0


import numpy as np
from .._supported_operators import sklearn_operator_name_map
from ..common.data_types import Int64TensorType
from ..common._apply_operation import apply_cast, apply_concat, apply_div, apply_reshape
from ..common._registration import register_converter
from ..common._topology import Scope, Operator
from ..common._container import ModelComponentContainer
from ..proto import onnx_proto


def _calculate_proba(scope, operator, container, model):
    """
    This function calculates class probability scores for
    BaggingClassifier.
    """
    final_proba_name = operator.outputs[1].full_name
    proba_list = []
    options = container.get_options(model, dict(raw_scores=False))
    use_raw_scores = options["raw_scores"]
    has_proba = hasattr(model.estimators_[0], "predict_proba") or (
        use_raw_scores and hasattr(model.estimators_[0], "decision_function")
    )
    for index, estimator in enumerate(model.estimators_):
        op_type = sklearn_operator_name_map[type(estimator)]

        this_operator = scope.declare_local_operator(op_type, estimator)
        if container.has_options(estimator, "raw_scores"):
            container.add_options(id(estimator), {"raw_scores": use_raw_scores})
            scope.add_options(id(estimator), {"raw_scores": use_raw_scores})

        label_name = scope.declare_local_variable("label_%d" % index, Int64TensorType())
        proba_name = scope.declare_local_variable(
            "proba_%d" % index, operator.inputs[0].type.__class__()
        )

        features = model.estimators_features_[index]
        n_features = (
            model.n_features_in_
            if hasattr(model, "n_features_in_")
            else model.n_features_
        )
        if len(features) == n_features and list(features) == list(range(n_features)):
            this_operator.inputs = operator.inputs
        else:
            # subset of features
            feat_name = scope.declare_local_variable(
                "fsel_%d" % index, operator.inputs[0].type.__class__()
            )
            index_name = scope.get_unique_variable_name("index_name_%d" % index)
            container.add_initializer(
                index_name,
                onnx_proto.TensorProto.INT64,
                (len(features),),
                list(features),
            )
            container.add_node(
                "Gather",
                [operator.inputs[0].full_name, index_name],
                [feat_name.full_name],
                name=scope.get_unique_operator_name("GatherBG"),
                axis=1,
            )
            this_operator.inputs.append(feat_name)

        this_operator.outputs.append(label_name)
        this_operator.outputs.append(proba_name)
        proba_output_name = proba_name.onnx_name if has_proba else label_name.onnx_name
        reshape_dim_val = len(model.classes_) if has_proba else 1
        reshaped_proba_name = scope.get_unique_variable_name("reshaped_proba")
        apply_reshape(
            scope,
            proba_output_name,
            reshaped_proba_name,
            container,
            desired_shape=(1, -1, reshape_dim_val),
        )
        proba_list.append(reshaped_proba_name)
    merged_proba_name = scope.get_unique_variable_name("merged_proba")
    apply_concat(scope, proba_list, merged_proba_name, container, axis=0)
    if has_proba:
        if container.target_opset >= 18:
            axis_name = scope.get_unique_variable_name("axis")
            container.add_initializer(axis_name, onnx_proto.TensorProto.INT64, [1], [0])
            container.add_node(
                "ReduceMean",
                [merged_proba_name, axis_name],
                final_proba_name,
                name=scope.get_unique_operator_name("ReduceMean"),
                keepdims=0,
            )
        else:
            container.add_node(
                "ReduceMean",
                merged_proba_name,
                final_proba_name,
                name=scope.get_unique_operator_name("ReduceMean"),
                axes=[0],
                keepdims=0,
            )
    else:
        n_estimators_name = scope.get_unique_variable_name("n_estimators")
        class_labels_name = scope.get_unique_variable_name("class_labels")
        equal_result_name = scope.get_unique_variable_name("equal_result")
        cast_output_name = scope.get_unique_variable_name("cast_output")
        reduced_proba_name = scope.get_unique_variable_name("reduced_proba")

        container.add_initializer(
            n_estimators_name,
            onnx_proto.TensorProto.FLOAT,
            [],
            [len(model.estimators_)],
        )
        container.add_initializer(
            class_labels_name,
            onnx_proto.TensorProto.INT64,
            [1, 1, len(model.estimators_[0].classes_)],
            model.estimators_[0].classes_,
        )

        container.add_node(
            "Equal",
            [class_labels_name, merged_proba_name],
            equal_result_name,
            name=scope.get_unique_operator_name("Equal"),
        )
        apply_cast(
            scope,
            equal_result_name,
            cast_output_name,
            container,
            to=onnx_proto.TensorProto.FLOAT,
        )
        if container.target_opset < 13:
            container.add_node(
                "ReduceSum",
                cast_output_name,
                reduced_proba_name,
                name=scope.get_unique_operator_name("ReduceSum"),
                axes=[0],
                keepdims=0,
            )
        else:
            axis_name = scope.get_unique_variable_name("axis")
            container.add_initializer(axis_name, onnx_proto.TensorProto.INT64, [1], [1])
            container.add_node(
                "ReduceSum",
                [cast_output_name, axis_name],
                reduced_proba_name,
                keepdims=0,
                name=scope.get_unique_operator_name("ReduceSum"),
            )
        apply_div(
            scope,
            [reduced_proba_name, n_estimators_name],
            final_proba_name,
            container,
            broadcast=1,
        )
    return final_proba_name


def convert_sklearn_bagging_classifier(
    scope: Scope, operator: Operator, container: ModelComponentContainer
):
    """
    Converter for BaggingClassifier.
    """
    if scope.get_options(operator.raw_operator, dict(nocl=False))["nocl"]:
        raise RuntimeError(
            "Option 'nocl' is not implemented for operator '{}'.".format(
                operator.raw_operator.__class__.__name__
            )
        )

    bagging_op = operator.raw_operator
    classes = bagging_op.classes_
    output_shape = (-1,)
    classes_name = scope.get_unique_variable_name("classes")
    argmax_output_name = scope.get_unique_variable_name("argmax_output")
    array_feature_extractor_result_name = scope.get_unique_variable_name(
        "array_feature_extractor_result"
    )
    class_type = onnx_proto.TensorProto.STRING

    if (
        np.issubdtype(bagging_op.classes_.dtype, np.floating)
        or bagging_op.classes_.dtype == np.bool_
    ):
        class_type = onnx_proto.TensorProto.INT32
        classes = classes.astype(np.int32)
    elif np.issubdtype(bagging_op.classes_.dtype, np.signedinteger):
        class_type = onnx_proto.TensorProto.INT32
    else:
        classes = np.array([s.encode("utf-8") for s in classes])

    container.add_initializer(classes_name, class_type, classes.shape, classes)

    proba_name = _calculate_proba(scope, operator, container, bagging_op)
    container.add_node(
        "ArgMax",
        proba_name,
        argmax_output_name,
        name=scope.get_unique_operator_name("ArgMax"),
        axis=1,
    )
    container.add_node(
        "ArrayFeatureExtractor",
        [classes_name, argmax_output_name],
        array_feature_extractor_result_name,
        op_domain="ai.onnx.ml",
        name=scope.get_unique_operator_name("ArrayFeatureExtractor"),
    )

    if class_type == onnx_proto.TensorProto.INT32:
        cast_result_name = scope.get_unique_variable_name("cast_result")
        reshaped_result_name = scope.get_unique_variable_name("reshaped_result")
        apply_cast(
            scope,
            array_feature_extractor_result_name,
            cast_result_name,
            container,
            to=onnx_proto.TensorProto.INT64,
        )
        apply_reshape(
            scope,
            cast_result_name,
            reshaped_result_name,
            container,
            desired_shape=output_shape,
        )
        apply_cast(
            scope,
            reshaped_result_name,
            operator.outputs[0].full_name,
            container,
            to=onnx_proto.TensorProto.INT64,
        )
    else:  # string labels
        apply_reshape(
            scope,
            array_feature_extractor_result_name,
            operator.outputs[0].full_name,
            container,
            desired_shape=output_shape,
        )


def convert_sklearn_bagging_regressor(
    scope: Scope, operator: Operator, container: ModelComponentContainer
):
    """
    Converter for BaggingRegressor.
    """
    bagging_op = operator.raw_operator
    proba_list = []
    for index, estimator in enumerate(bagging_op.estimators_):
        op_type = sklearn_operator_name_map[type(estimator)]
        this_operator = scope.declare_local_operator(op_type, estimator)

        features = bagging_op.estimators_features_[index]
        n_features = (
            bagging_op.n_features_in_
            if hasattr(bagging_op, "n_features_in_")
            else bagging_op.n_features_
        )
        if len(features) == n_features and list(features) == list(range(n_features)):
            this_operator.inputs = operator.inputs
        else:
            # subset of features
            feat_name = scope.declare_local_variable(
                "fsel_%d" % index, operator.inputs[0].type.__class__()
            )
            index_name = scope.get_unique_variable_name("index_name")
            container.add_initializer(
                index_name,
                onnx_proto.TensorProto.INT64,
                (len(features),),
                list(features),
            )
            container.add_node(
                "Gather",
                [operator.inputs[0].full_name, index_name],
                [feat_name.full_name],
                name=scope.get_unique_operator_name("GatherBG"),
                axis=1,
            )
            this_operator.inputs.append(feat_name)

        label_name = scope.declare_local_variable(
            "variable_%d" % index, this_operator.inputs[0].type.__class__()
        )
        this_operator.outputs.append(label_name)
        reshaped_proba_name = scope.get_unique_variable_name("reshaped_proba")
        apply_reshape(
            scope,
            label_name.onnx_name,
            reshaped_proba_name,
            container,
            desired_shape=(1, -1, 1),
        )
        proba_list.append(reshaped_proba_name)

    merged_proba_name = scope.get_unique_variable_name("merged_proba")
    apply_concat(scope, proba_list, merged_proba_name, container, axis=0)
    if container.target_opset >= 18:
        axis_name = scope.get_unique_variable_name("axis")
        container.add_initializer(axis_name, onnx_proto.TensorProto.INT64, [1], [0])
        container.add_node(
            "ReduceMean",
            [merged_proba_name, axis_name],
            operator.outputs[0].full_name,
            name=scope.get_unique_operator_name("ReduceMean"),
            keepdims=0,
        )
    else:
        container.add_node(
            "ReduceMean",
            merged_proba_name,
            operator.outputs[0].full_name,
            name=scope.get_unique_operator_name("ReduceMean"),
            axes=[0],
            keepdims=0,
        )


register_converter(
    "SklearnBaggingClassifier",
    convert_sklearn_bagging_classifier,
    options={
        "zipmap": [True, False, "columns"],
        "nocl": [True, False],
        "output_class_labels": [False, True],
        "raw_scores": [True, False],
    },
)
register_converter("SklearnBaggingRegressor", convert_sklearn_bagging_regressor)



# SPDX-License-Identifier: Apache-2.0

import numpy as np
from ..common._registration import register_converter
from ..common.data_types import guess_numpy_type
from ..common._topology import Scope, Operator
from ..common._container import ModelComponentContainer
from ..algebra.onnx_ops import OnnxMatMul


def convert_random_projection(
    scope: Scope, operator: Operator, container: ModelComponentContainer
):
    """Converter for PowerTransformer"""
    op_in = operator.inputs[0]
    op_out = operator.outputs[0].full_name
    op = operator.raw_operator
    opv = container.target_opset
    dtype = guess_numpy_type(operator.inputs[0].type)
    if dtype != np.float64:
        dtype = np.float32

    y = OnnxMatMul(
        op_in, op.components_.T.astype(dtype), op_version=opv, output_names=[op_out]
    )
    y.add_to(scope, container)


register_converter("SklearnGaussianRandomProjection", convert_random_projection)



# SPDX-License-Identifier: Apache-2.0

import warnings
import numpy as np

from sklearn import pipeline
from sklearn.base import ClassifierMixin, ClusterMixin, is_classifier

try:
    from sklearn.base import OutlierMixin
except ImportError:
    # scikit-learn <= 0.19
    class OutlierMixin:
        pass


from sklearn.ensemble import (
    IsolationForest,
    RandomTreesEmbedding,
    RandomForestClassifier,
)
from sklearn.gaussian_process import GaussianProcessRegressor
from sklearn.linear_model import BayesianRidge
from sklearn.model_selection import GridSearchCV
from sklearn.neighbors import NearestNeighbors, LocalOutlierFactor
from sklearn.mixture import GaussianMixture, BayesianGaussianMixture
from sklearn.multioutput import MultiOutputClassifier
from sklearn.preprocessing import OneHotEncoder, OrdinalEncoder
from sklearn.pipeline import Pipeline
from sklearn.svm import LinearSVC, NuSVC, SVC

try:
    from sklearn.compose import ColumnTransformer
except ImportError:
    # ColumnTransformer was introduced in 0.20.
    ColumnTransformer = None
try:
    from sklearn.preprocessing import Imputer
except ImportError:
    Imputer = None
try:
    from sklearn.impute import SimpleImputer
except ImportError:
    # changed in 0.20
    SimpleImputer = None

from ._supported_operators import _get_sklearn_operator_name, cluster_list, outlier_list
from ._supported_operators import sklearn_classifier_list, sklearn_operator_name_map
from .common._container import SklearnModelContainerNode
from .common._registration import _converter_pool, _shape_calculator_pool
from .common._topology import Topology, Variable
from .common.data_types import (
    DictionaryType,
    Int64TensorType,
    SequenceType,
    StringTensorType,
    TensorType,
    FloatTensorType,
    guess_tensor_type,
)
from .common.utils import get_column_indices
from .common.utils_checking import check_signature
from .common.utils_classifier import get_label_classes
from .common.utils_sklearn import _process_options


do_not_merge_columns = tuple(
    filter(
        lambda op: op is not None, [OrdinalEncoder, OneHotEncoder, ColumnTransformer]
    )
)


def _fetch_input_slice(scope, inputs, column_indices):
    if not isinstance(inputs, list):
        raise TypeError("Parameter inputs must be a list.")
    if len(inputs) == 0:
        raise RuntimeError(
            "Operator ArrayFeatureExtractor requires at " "least one inputs."
        )
    if len(inputs) != 1:
        raise RuntimeError(
            "Operator ArrayFeatureExtractor does not support " "multiple input tensors."
        )
    if (
        isinstance(inputs[0].type, TensorType)
        and len(inputs[0].type.shape) == 2
        and inputs[0].type.shape[1] == len(column_indices)
    ):
        # No need to extract.
        return inputs
    array_feature_extractor_operator = scope.declare_local_operator(
        "SklearnArrayFeatureExtractor"
    )
    array_feature_extractor_operator.inputs = inputs
    array_feature_extractor_operator.column_indices = column_indices
    output_variable_name = scope.declare_local_variable(
        "extracted_feature_columns", inputs[0].type
    )
    array_feature_extractor_operator.outputs.append(output_variable_name)
    return array_feature_extractor_operator.outputs


def _parse_sklearn_simple_model(scope, model, inputs, custom_parsers=None, alias=None):
    """
    This function handles all non-pipeline models.

    :param scope: Scope object
    :param model: A scikit-learn object (e.g., *OneHotEncoder*
        or *LogisticRegression*)
    :param inputs: A list of variables
    :param custom_parsers: dictionary of custom parsers
    :param alias: use this alias instead of the one based on the class name
    :return: A list of output variables which will be passed to next
        stage
    """
    # alias can be None
    if isinstance(model, str):
        raise RuntimeError(
            "Parameter model must be an object not a " "string '{0}'.".format(model)
        )
    if any(not isinstance(i, Variable) for i in inputs):
        raise TypeError(
            "One input is not a Variable for model %r - %r." "" % (model, inputs)
        )
    if alias is None:
        alias = _get_sklearn_operator_name(type(model))
    this_operator = scope.declare_local_operator(alias, model)
    this_operator.inputs = inputs

    if hasattr(model, "onnx_parser"):
        parser_names = model.onnx_parser()
        if parser_names is not None:
            try:
                names = parser_names(scope=scope, inputs=inputs)
            except TypeError as e:
                warnings.warn(
                    "Calling parser %r for model type %r failed due to %r. "
                    "This warnings will become an exception in version 1.11. "
                    "The parser signature should parser(scope=None, "
                    "inputs=None)." % (parser_names, e, type(model)),
                    DeprecationWarning,
                )
                names = parser_names()
            if names is not None:
                for name in names:
                    if isinstance(name, Variable):
                        this_operator.outputs.append(name)
                    elif isinstance(name, str):
                        var = scope.declare_local_variable(
                            name, guess_tensor_type(inputs[0].type)
                        )
                        this_operator.outputs.append(var)
                    elif isinstance(name, tuple) and len(name) == 2:
                        var = scope.declare_local_variable(
                            name[0], guess_tensor_type(name[1])
                        )
                        this_operator.outputs.append(var)
                    else:
                        raise RuntimeError(
                            "Unexpected output type %r (value=%r) for "
                            "operator %r." % (type(name), name, type(model))
                        )
                return this_operator.outputs

    if (
        type(model) in sklearn_classifier_list
        or isinstance(model, ClassifierMixin)
        or (isinstance(model, GridSearchCV) and is_classifier(model))
    ):
        # For classifiers, we may have two outputs, one for label and
        # the other one for probabilities of all classes. Notice that
        # their types here are not necessarily correct and they will
        # be fixed in shape inference phase.
        label_variable = scope.declare_local_variable("label", Int64TensorType())
        if type(model) in [RandomForestClassifier]:
            prob_dtype = FloatTensorType()
        else:
            prob_dtype = guess_tensor_type(inputs[0].type)
        probability_tensor_variable = scope.declare_local_variable(
            "probabilities", prob_dtype
        )
        this_operator.outputs.append(label_variable)
        this_operator.outputs.append(probability_tensor_variable)

    elif type(model) in cluster_list or isinstance(model, ClusterMixin):
        # For clustering, we may have two outputs, one for label and
        # the other one for scores of all classes. Notice that their
        # types here are not necessarily correct and they will be fixed
        # in shape inference phase
        label_variable = scope.declare_local_variable("label", Int64TensorType())
        score_tensor_variable = scope.declare_local_variable(
            "scores", guess_tensor_type(inputs[0].type)
        )
        this_operator.outputs.append(label_variable)
        this_operator.outputs.append(score_tensor_variable)

    elif type(model) in {IsolationForest, LocalOutlierFactor}:
        label_variable = scope.declare_local_variable("label", Int64TensorType())
        score_tensor_variable = scope.declare_local_variable(
            "scores", guess_tensor_type(inputs[0].type)
        )
        this_operator.outputs.append(label_variable)
        this_operator.outputs.append(score_tensor_variable)
        options = scope.get_options(model, dict(score_samples=False))
        if options["score_samples"]:
            scores_var = scope.declare_local_variable(
                "score_samples", guess_tensor_type(inputs[0].type)
            )
            this_operator.outputs.append(scores_var)

    elif type(model) in outlier_list or isinstance(model, OutlierMixin):
        # For outliers, we may have two outputs, one for label and
        # the other one for scores.
        label_variable = scope.declare_local_variable("label", Int64TensorType())
        score_tensor_variable = scope.declare_local_variable(
            "scores", guess_tensor_type(inputs[0].type)
        )
        this_operator.outputs.append(label_variable)
        this_operator.outputs.append(score_tensor_variable)

    elif isinstance(model, NearestNeighbors):
        # For Nearest Neighbours, we have two outputs, one for nearest
        # neighbours' indices and the other one for distances
        index_variable = scope.declare_local_variable("index", Int64TensorType())
        distance_variable = scope.declare_local_variable(
            "distance", guess_tensor_type(inputs[0].type)
        )
        this_operator.outputs.append(index_variable)
        this_operator.outputs.append(distance_variable)

    elif type(model) in {GaussianMixture, BayesianGaussianMixture}:
        label_variable = scope.declare_local_variable("label", Int64TensorType())
        prob_variable = scope.declare_local_variable(
            "probabilities", guess_tensor_type(inputs[0].type)
        )
        this_operator.outputs.append(label_variable)
        this_operator.outputs.append(prob_variable)
        options = scope.get_options(model, dict(score_samples=False))
        if options["score_samples"]:
            scores_var = scope.declare_local_variable(
                "score_samples", guess_tensor_type(inputs[0].type)
            )
            this_operator.outputs.append(scores_var)
    elif type(model) in {SimpleImputer, Imputer}:
        if isinstance(inputs[0].type, (Int64TensorType, StringTensorType)):
            otype = inputs[0].type.__class__()
        else:
            otype = guess_tensor_type(inputs[0].type)
        variable = scope.declare_local_variable("variable", otype)
        this_operator.outputs.append(variable)
    else:
        if hasattr(model, "get_feature_names_out"):
            try:
                out_names = model.get_feature_names_out()
            except (AttributeError, ValueError):
                # Catch a bug in scikit-learn.
                out_names = None
            this_operator.feature_names_out_ = out_names
        input_type = guess_tensor_type(inputs[0].type)
        variable = scope.declare_local_variable("variable", input_type)
        this_operator.outputs.append(variable)

    options = scope.get_options(model, dict(decision_path=False), fail=False)
    if options is not None and options["decision_path"]:
        dec_path = scope.declare_local_variable("decision_path", StringTensorType())
        this_operator.outputs.append(dec_path)

    options = scope.get_options(model, dict(decision_leaf=False), fail=False)
    if options is not None and options["decision_leaf"]:
        dec_path = scope.declare_local_variable("decision_leaf", Int64TensorType())
        this_operator.outputs.append(dec_path)

    return this_operator.outputs


def _parse_sklearn_pipeline(scope, model, inputs, custom_parsers=None):
    """
    The basic ideas of scikit-learn parsing:
        1. Sequentially go though all stages defined in the considered
           scikit-learn pipeline
        2. The output variables of one stage will be fed into its next
           stage as the inputs.

    :param scope: Scope object defined in _topology.py
    :param model: scikit-learn pipeline object
    :param inputs: A list of Variable objects
    :return: A list of output variables produced by the input pipeline
    """
    for step in model.steps:
        inputs = _parse_sklearn(scope, step[1], inputs, custom_parsers=custom_parsers)
    return inputs


def _parse_sklearn_feature_union(scope, model, inputs, custom_parsers=None):
    """
    :param scope: Scope object
    :param model: A scikit-learn FeatureUnion object
    :param inputs: A list of Variable objects
    :return: A list of output variables produced by feature union
    """
    # Output variable name of each transform. It's a list of string.
    transformed_result_names = []
    # Encode each transform as our IR object
    for name, transform in model.transformer_list:
        transformed_result_names.append(
            _parse_sklearn(scope, transform, inputs, custom_parsers=custom_parsers)[0]
        )
        if model.transformer_weights is not None and name in model.transformer_weights:
            transform_result = [transformed_result_names.pop()]
            # Create a Multiply ONNX node
            multiply_operator = scope.declare_local_operator("SklearnMultiply")
            multiply_operator.inputs = transform_result
            multiply_operator.operand = model.transformer_weights[name]
            multiply_output = scope.declare_local_variable(
                "multiply_output", guess_tensor_type(inputs[0].type)
            )
            multiply_operator.outputs.append(multiply_output)
            transformed_result_names.append(multiply_operator.outputs[0])

    # Create a Concat ONNX node
    concat_operator = scope.declare_local_operator("SklearnConcat")
    concat_operator.inputs = transformed_result_names

    # Declare output name of scikit-learn FeatureUnion
    union_name = scope.declare_local_variable(
        "union", guess_tensor_type(inputs[0].type)
    )
    concat_operator.outputs.append(union_name)

    return concat_operator.outputs


def _parse_sklearn_column_transformer(scope, model, inputs, custom_parsers=None):
    """
    :param scope: Scope object
    :param model: A *scikit-learn* *ColumnTransformer* object
    :param inputs: A list of Variable objects
    :return: A list of output variables produced by column transformer
    """
    # Output variable name of each transform. It's a list of string.
    transformed_result_names = []
    # Encode each transform as our IR object
    for name, op, column_indices in model.transformers_:
        if op == "drop":
            continue
        if isinstance(column_indices, slice):
            column_indices = list(
                range(
                    column_indices.start if column_indices.start is not None else 0,
                    column_indices.stop,
                    column_indices.step if column_indices.step is not None else 1,
                )
            )
        elif isinstance(column_indices, (int, str)):
            column_indices = [column_indices]
        names = get_column_indices(column_indices, inputs, multiple=True)
        transform_inputs = []
        for onnx_var, onnx_is in names.items():
            tr_inputs = _fetch_input_slice(scope, [inputs[onnx_var]], onnx_is)
            transform_inputs.extend(tr_inputs)

        merged_cols = False
        if len(transform_inputs) > 1:
            if isinstance(op, Pipeline):
                if not isinstance(op.steps[0][1], do_not_merge_columns):
                    merged_cols = True
            elif not isinstance(op, do_not_merge_columns):
                merged_cols = True

        if merged_cols:
            # Many ONNX operators expect one input vector,
            # the default behaviour is to merge columns.
            ty = transform_inputs[0].type.__class__([None, None])

            conc_op = scope.declare_local_operator("SklearnConcat")
            conc_op.inputs = transform_inputs
            conc_names = scope.declare_local_variable("merged_columns", ty)
            conc_op.outputs.append(conc_names)
            transform_inputs = [conc_names]

        model_obj = model.named_transformers_[name]
        if isinstance(model_obj, str):
            if model_obj == "passthrough":
                var_out = transform_inputs[0]
            elif model_obj == "drop":
                var_out = None
            else:
                raise RuntimeError(
                    "Unknown operator alias "
                    "'{0}'. These are specified in "
                    "_supported_operators.py."
                    "".format(model_obj)
                )
        else:
            var_out = _parse_sklearn(
                scope, model_obj, transform_inputs, custom_parsers=custom_parsers
            )[0]
            if (
                model.transformer_weights is not None
                and name in model.transformer_weights
            ):
                # Create a Multiply ONNX node
                multiply_operator = scope.declare_local_operator("SklearnMultiply")
                multiply_operator.inputs.append(var_out)
                multiply_operator.operand = model.transformer_weights[name]
                var_out = scope.declare_local_variable(
                    "multiply_output", guess_tensor_type(inputs[0].type)
                )
                multiply_operator.outputs.append(var_out)
        if var_out:
            transformed_result_names.append(var_out)

    # Create a Concat ONNX node
    if len(transformed_result_names) > 1:
        ty = transformed_result_names[0].type.__class__([None, None])
        concat_operator = scope.declare_local_operator("SklearnConcat")
        concat_operator.inputs = transformed_result_names

        # Declare output name of scikit-learn ColumnTransformer
        transformed_column_name = scope.declare_local_variable("transformed_column", ty)
        concat_operator.outputs.append(transformed_column_name)
        return concat_operator.outputs
    return transformed_result_names


def _parse_sklearn_grid_search_cv(scope, model, inputs, custom_parsers=None):
    options = scope.get_options(model)
    if options:
        scope.add_options(id(model.best_estimator_), options)
    res = parse_sklearn(
        scope, model.best_estimator_, inputs, custom_parsers=custom_parsers
    )
    scope.replace_raw_operator(model.best_estimator_, model, "SklearnGridSearchCV")
    return res


def _parse_sklearn_random_trees_embedding(scope, model, inputs, custom_parsers=None):
    res = parse_sklearn(
        scope, model.base_estimator_, inputs, custom_parsers=custom_parsers
    )
    if len(res) != 1:
        raise RuntimeError("A regressor only produces one output not %r." % res)
    scope.replace_raw_operator(
        model.base_estimator_, model, "SklearnRandomTreesEmbedding"
    )
    return res


def _apply_zipmap(zipmap_options, scope, model, input_type, probability_tensor):
    if zipmap_options == "columns":
        zipmap_operator = scope.declare_local_operator("SklearnZipMapColumns")
        classes = get_label_classes(scope, model)
        classes_names = get_label_classes(scope, model, node_names=True)
    else:
        zipmap_operator = scope.declare_local_operator("SklearnZipMap")
        classes = get_label_classes(scope, model)

    zipmap_operator.inputs = probability_tensor
    label_type = Int64TensorType([None])

    if (
        hasattr(model, "classes_")
        and isinstance(model.classes_, list)
        and isinstance(model.classes_[0], np.ndarray)
    ):
        # multi-label problem
        pass
    elif np.issubdtype(classes.dtype, np.floating):
        classes = np.array(list(map(lambda x: int(x), classes)))
        if set(map(lambda x: float(x), classes)) != set(model.classes_):
            raise RuntimeError(
                "skl2onnx implicitly converts float class "
                "labels into integers but at least one label "
                "is not an integer. Class labels should "
                "be integers or strings."
            )
        zipmap_operator.classlabels_int64s = classes
    elif np.issubdtype(classes.dtype, np.signedinteger):
        zipmap_operator.classlabels_int64s = [int(i) for i in classes]
    elif np.issubdtype(classes.dtype, np.unsignedinteger) or classes.dtype == np.bool_:
        zipmap_operator.classlabels_int64s = [int(i) for i in classes]
    else:
        classes = np.array([s.encode("utf-8") for s in classes])
        zipmap_operator.classlabels_strings = classes
        label_type = StringTensorType([None])

    zip_label = scope.declare_local_variable("output_label", label_type)
    if len(probability_tensor) == 2:
        zipmap_operator.outputs.append(zip_label)

    if zipmap_options == "columns":
        prob_type = probability_tensor[-1].type
        for cl in classes_names:
            output_cl = scope.declare_local_variable(cl, prob_type.__class__())
            zipmap_operator.outputs.append(output_cl)
    else:
        zip_probability = scope.declare_local_variable(
            "output_probability",
            SequenceType(DictionaryType(label_type, guess_tensor_type(input_type))),
        )
        zipmap_operator.outputs.append(zip_probability)

    zipmap_operator.init_status(is_evaluated=True)
    return zipmap_operator.outputs


def _parse_sklearn_classifier(scope, model, inputs, custom_parsers=None):
    options = scope.get_options(model, dict(zipmap=True))
    no_zipmap = (isinstance(options["zipmap"], bool) and not options["zipmap"]) or (
        model.__class__ in [NuSVC, SVC] and not model.probability
    )
    probability_tensor = _parse_sklearn_simple_model(
        scope, model, inputs, custom_parsers=custom_parsers
    )

    if no_zipmap:
        if options.get("output_class_labels", False):
            if not hasattr(model, "classes_"):
                raise RuntimeError(
                    "Model type %r has no attribute 'classes_'. "
                    "Option 'output_class_labels' is invalid or a new parser "
                    "must be used." % model.__class__.__name__
                )

            clout = scope.declare_local_operator("SklearnClassLabels")
            clout.classes = get_label_classes(scope, model)
            if model.classes_.dtype in (np.int32, np.int64, np.bool_):
                ctype = Int64TensorType
            else:
                ctype = StringTensorType
            label_type = ctype(clout.classes.shape)
            class_labels = scope.declare_local_variable("class_labels", label_type)
            clout.outputs.append(class_labels)
            outputs = list(probability_tensor)
            outputs.append(class_labels)
            return outputs
        return probability_tensor

    if options.get("output_class_labels", False):
        raise RuntimeError(
            "Option 'output_class_labels' is not compatible with option " "'zipmap'."
        )

    return _apply_zipmap(
        options["zipmap"], scope, model, inputs[0].type, probability_tensor
    )


def _parse_sklearn_multi_output_classifier(scope, model, inputs, custom_parsers=None):
    options = scope.get_options(model, dict(zipmap=True))
    if options["zipmap"]:
        warnings.warn(
            "Option zipmap is ignored for model %r. "
            "Set option zipmap to False to "
            "remove this message." % type(model),
            UserWarning,
        )
    alias = _get_sklearn_operator_name(type(model))
    this_operator = scope.declare_local_operator(alias, model)
    this_operator.inputs = inputs

    if hasattr(model, "classes_"):
        classes = model.classes_
    else:
        classes = [get_label_classes(scope, m) for m in model.estimators_]
    if len(set(cl.dtype for cl in classes)) != 1:
        raise RuntimeError(
            "Class labels may have only one type %r."
            "" % set(cl.dtype for cl in classes)
        )
    if classes[0].dtype in (np.int32, np.int64, np.bool_):
        ctype = Int64TensorType
    else:
        ctype = StringTensorType

    label = scope.declare_local_variable("label", ctype())
    proba = scope.declare_local_variable(
        "probabilities", SequenceType(guess_tensor_type(inputs[0].type))
    )
    this_operator.outputs.append(label)
    this_operator.outputs.append(proba)

    options = scope.get_options(model)
    if options.get("output_class_labels", False):
        clout = scope.declare_local_operator("SklearnClassLabels")
        clout.is_multi_output = True
        clout.classes = classes
        class_labels = scope.declare_local_variable(
            "class_labels", SequenceType(ctype())
        )
        clout.outputs.append(class_labels)
        return list(this_operator.outputs) + [class_labels]

    return this_operator.outputs


def _parse_sklearn_gaussian_process(scope, model, inputs, custom_parsers=None):
    options = scope.get_options(model, dict(return_cov=False, return_std=False))
    if options["return_std"] and options["return_cov"]:
        raise RuntimeError(
            "Not returning standard deviation of predictions when "
            "returning full covariance."
        )

    alias = _get_sklearn_operator_name(type(model))
    this_operator = scope.declare_local_operator(alias, model)
    mean_tensor = scope.declare_local_variable(
        "GPmean", guess_tensor_type(inputs[0].type)
    )
    this_operator.inputs = inputs
    this_operator.outputs.append(mean_tensor)

    if options["return_std"] or options["return_cov"]:
        # covariance or standard deviation
        covstd_tensor = scope.declare_local_variable(
            "GPcovstd", guess_tensor_type(inputs[0].type)
        )
        this_operator.outputs.append(covstd_tensor)
    return this_operator.outputs


def _parse_sklearn_bayesian_ridge(scope, model, inputs, custom_parsers=None):
    options = scope.get_options(model, dict(return_std=False))
    alias = _get_sklearn_operator_name(type(model))
    this_operator = scope.declare_local_operator(alias, model)
    mean_tensor = scope.declare_local_variable(
        "variable", guess_tensor_type(inputs[0].type)
    )
    this_operator.inputs = inputs
    this_operator.outputs.append(mean_tensor)

    if options["return_std"]:
        # covariance or standard deviation
        covstd_tensor = scope.declare_local_variable(
            "std", guess_tensor_type(inputs[0].type)
        )
        this_operator.outputs.append(covstd_tensor)
    return this_operator.outputs


def _parse_sklearn(scope, model, inputs, custom_parsers=None, alias=None):
    """
    This is a delegate function. It does nothing but invokes the
    correct parsing function according to the input model's type.

    :param scope: Scope object
    :param model: A scikit-learn object (e.g., OneHotEncoder
        and LogisticRegression)
    :param inputs: A list of variables
    :param custom_parsers: parsers determines which outputs is expected
        for which particular task, default parsers are defined for
        classifiers, regressors, pipeline but they can be rewritten,
        *custom_parsers* is a dictionary ``{ type: fct_parser(scope,
        model, inputs, custom_parsers=None) }``
    :param alias: alias of the model (None if based on the model class)
    :return: The output variables produced by the input model
    """
    for i, inp in enumerate(inputs):
        if not isinstance(inp, Variable):
            raise TypeError(
                "Unexpected input type %r for input %r: %r." % (type(inp), i, inp)
            )

    if alias is not None:
        outputs = _parse_sklearn_simple_model(
            scope, model, inputs, custom_parsers=custom_parsers, alias=alias
        )
        return outputs

    tmodel = type(model)
    if custom_parsers is not None and tmodel in custom_parsers:
        outputs = custom_parsers[tmodel](
            scope, model, inputs, custom_parsers=custom_parsers
        )
    elif tmodel in sklearn_parsers_map:
        outputs = sklearn_parsers_map[tmodel](
            scope, model, inputs, custom_parsers=custom_parsers
        )
    elif isinstance(model, pipeline.Pipeline):
        parser = sklearn_parsers_map[pipeline.Pipeline]
        outputs = parser(scope, model, inputs, custom_parsers=custom_parsers)
    else:
        outputs = _parse_sklearn_simple_model(
            scope, model, inputs, custom_parsers=custom_parsers
        )
    return outputs


def parse_sklearn(scope, model, inputs, custom_parsers=None, final_types=None):
    """
    This is a delegate function. It does nothing but invokes the
    correct parsing function according to the input model's type.

    :param scope: Scope object
    :param model: A scikit-learn object (e.g., OneHotEncoder
        and LogisticRegression)
    :param inputs: A list of variables
    :param custom_parsers: parsers determines which outputs is expected
        for which particular task, default parsers are defined for
        classifiers, regressors, pipeline but they can be rewritten,
        *custom_parsers* is a dictionary ``{ type: fct_parser(scope,
        model, inputs, custom_parsers=None) }``
    :param final_types: a python list. Works the same way as initial_types
        but not mandatory, it is used to overwrites the type
        (if type is not None) and the name of every output.
    :return: The output variables produced by the input model
    """
    if final_types is not None:
        outputs = []
        for name, ty in final_types:
            var = scope.declare_local_output(name, ty, missing_type=True)
            if var.onnx_name != name:
                raise RuntimeError(
                    "Unable to add duplicated output '{}', '{}'. "
                    "Output and input must have different names."
                    "".format(var.onnx_name, name)
                )
            outputs.append(var)

        hidden_outputs = _parse_sklearn(
            scope, model, inputs, custom_parsers=custom_parsers
        )

        if len(hidden_outputs) != len(outputs):
            raise RuntimeError(
                "Number of declared outputs is unexpected, declared '{}' "
                "found '{}'.".format(
                    ", ".join(_.onnx_name for _ in outputs),
                    ", ".join(_.onnx_name for _ in hidden_outputs),
                )
            )
        for h, o in zip(hidden_outputs, outputs):
            if o.type is None:
                iop = scope.declare_local_operator("SklearnIdentity")
            else:
                iop = scope.declare_local_operator("SklearnCast")
            iop.inputs = [h]
            iop.outputs = [o]
            h.init_status(is_leaf=False)
            o.init_status(is_leaf=True)
            if o.type is None and h.type is not None:
                o.type = h.type
        return outputs

    res = _parse_sklearn(scope, model, inputs, custom_parsers=custom_parsers)
    for r in res:
        r.init_status(is_leaf=True)
    return res


def parse_sklearn_model(
    model,
    initial_types=None,
    target_opset=None,
    custom_conversion_functions=None,
    custom_shape_calculators=None,
    custom_parsers=None,
    options=None,
    white_op=None,
    black_op=None,
    final_types=None,
    naming=None,
):
    """
    Puts *scikit-learn* object into an abstract container so that
    our framework can work seamlessly on models created
    with different machine learning tools.

    :param model: A scikit-learn model
    :param initial_types: a python list. Each element is a tuple of a
        variable name and a type defined in data_types.py
    :param target_opset: number, for example, 7 for ONNX 1.2,
        and 8 for ONNX 1.3.
    :param custom_conversion_functions: a dictionary for specifying
        the user customized conversion function if not registered
    :param custom_shape_calculators: a dictionary for specifying the
        user customized shape calculator if not registered
    :param custom_parsers: parsers determines which outputs is expected
        for which particular task, default parsers are defined for
        classifiers, regressors, pipeline but they can be rewritten,
        *custom_parsers* is a dictionary
        ``{ type: fct_parser(scope, model, inputs, custom_parsers=None) }``
    :param options: specific options given to converters
        (see :ref:`l-conv-options`)
    :param white_op: white list of ONNX nodes allowed
        while converting a pipeline, if empty, all are allowed
    :param black_op: black list of ONNX nodes allowed
        while converting a pipeline, if empty, none are blacklisted
    :param final_types: a python list. Works the same way as initial_types
        but not mandatory, it is used to overwrites the type
        (if type is not None) and the name of every output.
    :param naming: the user may want to change the way intermediate
        are named, this parameter can be a string (a prefix) or a
        function, which signature is the following:
        `get_name(name, existing_names)`, the library will then
        check this name is unique and modify it if not
    :return: :class:`Topology <skl2onnx.common._topology.Topology>`

    .. versionchanged:: 1.10.0
        Parameter *naming* was added.
    """
    options = _process_options(model, options)

    raw_model_container = SklearnModelContainerNode(
        model, white_op=white_op, black_op=black_op
    )

    # Declare a computational graph. It will become a representation of
    # the input scikit-learn model after parsing.
    topology = Topology(
        raw_model_container,
        initial_types=initial_types,
        target_opset=target_opset,
        custom_conversion_functions=custom_conversion_functions,
        custom_shape_calculators=custom_shape_calculators,
        registered_models=dict(
            conv=_converter_pool,
            shape=_shape_calculator_pool,
            aliases=sklearn_operator_name_map,
        ),
    )

    # Declare an object to provide variables' and operators' naming mechanism.
    scope = topology.declare_scope("__root__", options=options, naming=naming)
    inputs = scope.input_variables

    # The object raw_model_container is a part of the topology
    # we're going to return. We use it to store the inputs of
    # the scikit-learn's computational graph.
    for variable in inputs:
        variable.init_status(is_root=True)
        raw_model_container.add_input(variable)

    # Parse the input scikit-learn model as a Topology object.
    outputs = parse_sklearn(
        scope, model, inputs, custom_parsers=custom_parsers, final_types=final_types
    )

    # The object raw_model_container is a part of the topology we're
    # going to return. We use it to store the outputs of the
    # scikit-learn's computational graph.
    if final_types is not None and len(final_types) != len(outputs):
        raise RuntimeError(
            "Unexpected number of outputs, expected %d, got %d "
            "after parsing." % (len(final_types), len(outputs))
        )
    return topology


def build_sklearn_parsers_map():
    map_parser = {
        pipeline.Pipeline: _parse_sklearn_pipeline,
        pipeline.FeatureUnion: _parse_sklearn_feature_union,
        BayesianRidge: _parse_sklearn_bayesian_ridge,
        GaussianProcessRegressor: _parse_sklearn_gaussian_process,
        GridSearchCV: _parse_sklearn_grid_search_cv,
        MultiOutputClassifier: _parse_sklearn_multi_output_classifier,
        RandomTreesEmbedding: _parse_sklearn_random_trees_embedding,
    }
    if ColumnTransformer is not None:
        map_parser[ColumnTransformer] = _parse_sklearn_column_transformer

    for tmodel in sklearn_classifier_list:
        if tmodel not in [LinearSVC]:
            map_parser[tmodel] = _parse_sklearn_classifier
    return map_parser


def update_registered_parser(model, parser_fct):
    """
    Registers or updates a parser for a new model.
    A parser returns the expected output of a model.

    :param model: model class
    :param parser_fct: parser, signature is the same as
        :func:`parse_sklearn <skl2onnx._parse.parse_sklearn>`
    """
    check_signature(parser_fct, _parse_sklearn_classifier)
    sklearn_parsers_map[model] = parser_fct


# registered parsers
sklearn_parsers_map = build_sklearn_parsers_map()



# SPDX-License-Identifier: Apache-2.0


from ..proto import onnx_proto
from ..common.data_types import DoubleTensorType
from ..common._registration import register_converter
from ..common._topology import Scope, Operator
from ..common._container import ModelComponentContainer
from .common import concatenate_variables


def convert_sklearn_binarizer(
    scope: Scope, operator: Operator, container: ModelComponentContainer
):
    feature_name = concatenate_variables(scope, operator.inputs, container)

    if isinstance(operator.inputs[0].type, DoubleTensorType):
        name0 = scope.get_unique_variable_name("cst0")
        name1 = scope.get_unique_variable_name("cst1")
        thres = scope.get_unique_variable_name("th")
        container.add_initializer(name0, onnx_proto.TensorProto.DOUBLE, [], [0.0])
        container.add_initializer(name1, onnx_proto.TensorProto.DOUBLE, [], [1.0])
        container.add_initializer(
            thres,
            onnx_proto.TensorProto.DOUBLE,
            [],
            [float(operator.raw_operator.threshold)],
        )
        binbool = scope.get_unique_variable_name("binbool")
        container.add_node(
            "Less",
            [feature_name, thres],
            binbool,
            name=scope.get_unique_operator_name("Less"),
        )
        container.add_node(
            "Where", [binbool, name0, name1], operator.output_full_names, name="Where"
        )
        return

    op_type = "Binarizer"
    attrs = {
        "name": scope.get_unique_operator_name(op_type),
        "threshold": float(operator.raw_operator.threshold),
    }
    container.add_node(
        op_type,
        feature_name,
        operator.output_full_names,
        op_domain="ai.onnx.ml",
        **attrs
    )


register_converter("SklearnBinarizer", convert_sklearn_binarizer)



# SPDX-License-Identifier: Apache-2.0


from ..common._registration import register_shape_calculator
from ..common.shape_calculator import calculate_linear_classifier_output_shapes


register_shape_calculator(
    "SklearnLinearClassifier", calculate_linear_classifier_output_shapes
)
register_shape_calculator("SklearnLinearSVC", calculate_linear_classifier_output_shapes)
register_shape_calculator(
    "SklearnAdaBoostClassifier", calculate_linear_classifier_output_shapes
)
register_shape_calculator(
    "SklearnBaggingClassifier", calculate_linear_classifier_output_shapes
)
register_shape_calculator(
    "SklearnBernoulliNB", calculate_linear_classifier_output_shapes
)
register_shape_calculator(
    "SklearnCategoricalNB", calculate_linear_classifier_output_shapes
)
register_shape_calculator(
    "SklearnComplementNB", calculate_linear_classifier_output_shapes
)
register_shape_calculator(
    "SklearnGaussianNB", calculate_linear_classifier_output_shapes
)
register_shape_calculator(
    "SklearnMultinomialNB", calculate_linear_classifier_output_shapes
)
register_shape_calculator(
    "SklearnCalibratedClassifierCV", calculate_linear_classifier_output_shapes
)
register_shape_calculator(
    "SklearnMLPClassifier", calculate_linear_classifier_output_shapes
)
register_shape_calculator(
    "SklearnSGDClassifier", calculate_linear_classifier_output_shapes
)
register_shape_calculator(
    "SklearnStackingClassifier", calculate_linear_classifier_output_shapes
)



# SPDX-License-Identifier: Apache-2.0

from ..common._registration import register_converter
from ..common._topology import Scope, Operator
from ..common._container import ModelComponentContainer


def convert_sklearn_flatten(
    scope: Scope, operator: Operator, container: ModelComponentContainer
):
    name = scope.get_unique_operator_name("Flatten")
    container.add_node(
        "Flatten",
        operator.inputs[0].full_name,
        operator.outputs[0].full_name,
        name=name,
        axis=1,
    )


register_converter("SklearnFlatten", convert_sklearn_flatten)



# SPDX-License-Identifier: Apache-2.0


import numpy as np
from ..proto import onnx_proto
from ..common.data_types import guess_numpy_type, guess_proto_type
from ..common._registration import register_converter
from ..common._topology import Scope, Operator
from ..common._container import ModelComponentContainer
from ..common._apply_operation import (
    apply_add,
    apply_log,
    apply_mul,
    apply_identity,
    apply_normalizer,
)


def convert_sklearn_tfidf_transformer(
    scope: Scope, operator: Operator, container: ModelComponentContainer
):
    # TODO: use sparse containers when available
    dtype = guess_numpy_type(operator.inputs[0].type)
    if dtype != np.float64:
        dtype = np.float32
    float_type = dtype
    # onnx_proto.TensorProto.FLOAT
    proto_dtype = guess_proto_type(operator.inputs[0].type)
    if proto_dtype != onnx_proto.TensorProto.DOUBLE:
        proto_dtype = onnx_proto.TensorProto.FLOAT
    op = operator.raw_operator
    data = operator.input_full_names
    output_name = scope.get_unique_variable_name("tfidftr_output")

    if op.sublinear_tf:
        # code scikit-learn
        # np.log(X.data, X.data) --> does not apply on null coefficient
        # X.data += 1
        # ONNX does not support sparse tensors before opset < 11
        # approximated by X.data += 1 --> np.log(X.data, X.data)
        if operator.target_opset < 11:
            plus1 = scope.get_unique_variable_name("plus1")
            C = operator.inputs[0].type.shape[1]
            ones = scope.get_unique_variable_name("ones")
            cst = np.ones((C,), dtype=float_type)
            container.add_initializer(ones, proto_dtype, [C], cst.flatten())
            apply_add(scope, data + [ones], plus1, container, broadcast=1)
            plus1logged = scope.get_unique_variable_name("plus1logged")
            apply_log(scope, plus1, plus1logged, container)
            data = [plus1logged]
        else:
            # sparse containers have not yet been implemented.
            raise RuntimeError(
                "ONNX does not support sparse tensors before opset < 11, "
                "sublinear_tf must be False."
            )

    if op.use_idf:
        cst = op.idf_.astype(float_type)
        if len(cst.shape) > 1:
            cst = np.diag(cst)
        cst = cst.ravel().flatten()
        shape = [len(cst)]
        idfcst = scope.get_unique_variable_name("idfcst")
        container.add_initializer(idfcst, proto_dtype, shape, cst)
        apply_mul(scope, data + [idfcst], output_name, container, broadcast=1)
    else:
        output_name = data[0]

    if op.norm is not None:
        norm_name = scope.get_unique_variable_name("tfidftr_norm")
        apply_normalizer(
            scope,
            output_name,
            norm_name,
            container,
            norm=op.norm.upper(),
            use_float=float_type == np.float32,
        )
        output_name = norm_name

    options = container.get_options(op, dict(nan=False))
    replace_by_nan = options.get("nan", False)
    if replace_by_nan:
        # This part replaces all null values by nan.
        cst_nan_name = scope.get_unique_variable_name("nan_name")
        container.add_initializer(cst_nan_name, proto_dtype, [1], [np.nan])
        cst_zero_name = scope.get_unique_variable_name("zero_name")
        container.add_initializer(cst_zero_name, proto_dtype, [1], [0])

        mask_name = scope.get_unique_variable_name("mask_name")
        container.add_node(
            "Equal",
            [output_name, cst_zero_name],
            mask_name,
            name=scope.get_unique_operator_name("Equal"),
        )

        where_name = scope.get_unique_variable_name("where_name")
        container.add_node(
            "Where",
            [mask_name, cst_nan_name, output_name],
            where_name,
            name=scope.get_unique_operator_name("Where"),
        )
        output_name = where_name

    apply_identity(scope, output_name, operator.output_full_names, container)


register_converter(
    "SklearnTfidfTransformer",
    convert_sklearn_tfidf_transformer,
    options={"nan": [True, False]},
)



# SPDX-License-Identifier: Apache-2.0

from .. import update_registered_converter
from ..shape_calculators.text_vectorizer import (
    calculate_sklearn_text_vectorizer_output_shapes,
)
from ..operator_converters.text_vectoriser import convert_sklearn_text_vectorizer
from ..operator_converters.tfidf_vectoriser import convert_sklearn_tfidf_vectoriser
from .sklearn_text import TraceableCountVectorizer, TraceableTfidfVectorizer


def register():
    """Register converter for TraceableCountVectorizer,
    TraceableTfidfVectorizer."""
    update_registered_converter(
        TraceableCountVectorizer,
        "Skl2onnxTraceableCountVectorizer",
        calculate_sklearn_text_vectorizer_output_shapes,
        convert_sklearn_text_vectorizer,
        options={
            "tokenexp": None,
            "separators": None,
            "nan": [True, False],
            "keep_empty_string": [True, False],
            "locale": None,
        },
    )

    update_registered_converter(
        TraceableTfidfVectorizer,
        "Skl2onnxTraceableTfidfVectorizer",
        calculate_sklearn_text_vectorizer_output_shapes,
        convert_sklearn_tfidf_vectoriser,
        options={
            "tokenexp": None,
            "separators": None,
            "nan": [True, False],
            "keep_empty_string": [True, False],
            "locale": None,
        },
    )



# SPDX-License-Identifier: Apache-2.0

import numpy as np

try:
    from sklearn.ensemble._iforest import _average_path_length
except ImportError:
    # scikit-learn < 0.22
    from sklearn.ensemble.iforest import _average_path_length
from ..common._registration import register_converter
from ..common.data_types import (
    BooleanTensorType,
    Int64TensorType,
    guess_numpy_type,
    guess_proto_type,
)
from ..common.tree_ensemble import (
    add_tree_to_attribute_pairs,
    get_default_tree_regressor_attribute_pairs,
)
from ..proto import onnx_proto
from ..algebra.onnx_ops import (
    OnnxTreeEnsembleRegressor_1,
    OnnxLog,
    OnnxCast,
    OnnxLess,
    OnnxLabelEncoder,
    OnnxMul,
    OnnxGreater,
    OnnxAdd,
    OnnxDiv,
    OnnxSum,
    OnnxNeg,
    OnnxReshapeApi13,
    OnnxEqual,
    OnnxPow,
    OnnxGather,
    OnnxMax,
)


def convert_sklearn_isolation_forest(
    scope,
    operator,
    container,
    op_type="TreeEnsembleRegressor",
    op_domain="ai.onnx.ml",
    op_version=1,
):
    op = operator.raw_operator
    outputs = operator.outputs
    opv = container.target_opset
    opvml = container.target_opset_any_domain("ai.onnx.ml")
    options = container.get_options(op, dict(score_samples=None))
    if opvml < 2:
        raise RuntimeError(
            "This converter requires at least opset 2 for " "domain 'ai.onnx.ml'."
        )

    input_name = operator.inputs[0]
    dtype = guess_numpy_type(operator.inputs[0].type)
    proto_dtype = guess_proto_type(operator.inputs[0].type)
    if type(operator.inputs[0].type) in (BooleanTensorType, Int64TensorType):
        input_name = OnnxCast(input_name, to=proto_dtype, op_version=opv)

    if op._max_features != operator.inputs[0].type.shape[1]:
        raise RuntimeError(
            "Converter for IsolationForest does not support the case when "
            "_max_features={} != number of given features {}.".format(
                op._max_features, operator.inputs[0].type.shape[1]
            )
        )

    # decision_path
    scores = []
    for i, (tree, features) in enumerate(zip(op.estimators_, op.estimators_features_)):
        # X_subset = X[:, features]
        gather = OnnxGather(
            input_name, features.astype(np.int64), axis=1, op_version=opv
        )

        attrs = get_default_tree_regressor_attribute_pairs()
        attrs["n_targets"] = 1
        add_tree_to_attribute_pairs(
            attrs, False, tree.tree_, 0, 1.0, 0, False, True, dtype=dtype
        )

        # tree leave
        attrs["n_targets"] = 1
        attrs["post_transform"] = "NONE"
        attrs["target_ids"] = [0 for _ in attrs["target_ids"]]
        attrs["target_weights"] = [float(_) for _ in attrs["target_nodeids"]]
        leave = OnnxTreeEnsembleRegressor_1(gather, op_version=1, **attrs)

        # tree - retrieve node_sample
        labels = _build_labels(tree.tree_, output="node_sample")
        ordered = list(sorted(labels.items()))
        values = [float(_[1]) for _ in ordered]
        if any(map(lambda i: int(i[0]) != i[0], ordered)):
            keys = [float(_[0]) for _ in ordered]
            node_sample = OnnxReshapeApi13(
                OnnxLabelEncoder(
                    leave, op_version=opvml, keys_floats=keys, values_floats=values
                ),
                np.array([-1, 1], dtype=np.int64),
                op_version=opv,
            )
        else:
            keys = [int(_[0]) for _ in ordered]
            values = [float(_[1]) for _ in ordered]
            node_sample = OnnxReshapeApi13(
                OnnxLabelEncoder(
                    OnnxCast(leave, op_version=opv, to=onnx_proto.TensorProto.INT64),
                    op_version=opvml,
                    keys_int64s=keys,
                    values_floats=values,
                ),
                np.array([-1, 1], dtype=np.int64),
                op_version=opv,
            )
        node_sample.set_onnx_name_prefix("node_sample%d" % i)

        # tree - retrieve path_length
        labels = _build_labels(tree.tree_, output="path_length")
        ordered = list(sorted(labels.items()))
        values = [float(_[1]) for _ in ordered]
        if any(map(lambda i: int(i[0]) != i[0], ordered)):
            keys = [float(_[0]) for _ in ordered]
            values = [float(_[1]) for _ in ordered]
            path_length = OnnxReshapeApi13(
                OnnxLabelEncoder(
                    leave, op_version=opvml, keys_floats=keys, values_floats=values
                ),
                np.array([-1, 1], dtype=np.int64),
                op_version=opv,
            )
        else:
            keys = [int(_[0]) for _ in ordered]
            path_length = OnnxReshapeApi13(
                OnnxLabelEncoder(
                    OnnxCast(leave, op_version=opv, to=onnx_proto.TensorProto.INT64),
                    op_version=opvml,
                    keys_int64s=keys,
                    values_floats=values,
                ),
                np.array([-1, 1], dtype=np.int64),
                op_version=opv,
            )
        path_length.set_onnx_name_prefix("path_length%d" % i)

        # score
        eq2 = OnnxCast(
            OnnxEqual(node_sample, np.array([2], dtype=np.float32), op_version=opv),
            to=proto_dtype,
            op_version=opv,
        )
        eq2.set_onnx_name_prefix("eq2_%d" % i)

        # 2.0 * (np.log(n_samples_leaf[not_mask] - 1.0) + np.euler_gamma)

        eqp2p = OnnxCast(
            OnnxGreater(node_sample, np.array([2], dtype=np.float32), op_version=opv),
            to=proto_dtype,
            op_version=opv,
        )
        eqp2p.set_onnx_name_prefix("plus2_%d" % i)

        eqp2ps = OnnxMul(eqp2p, node_sample, op_version=opv)
        eqp2ps.set_onnx_name_prefix("eqp2ps%d" % i)

        eqp2ps_1 = OnnxAdd(eqp2ps, np.array([-1], dtype=dtype), op_version=opv)

        eqp2p_m1 = OnnxMax(eqp2ps_1, np.array([1], dtype=dtype), op_version=opv)
        eqp2p_m1.set_onnx_name_prefix("eqp2p_m1_%d" % i)

        eqp_log = OnnxMul(
            OnnxAdd(
                OnnxLog(eqp2p_m1, op_version=opv),
                np.array([np.euler_gamma], dtype=dtype),
                op_version=opv,
            ),
            np.array([2], dtype=dtype),
            op_version=opv,
        )
        eqp_log.set_onnx_name_prefix("eqp_log%d" % i)

        # - 2.0 * (n_samples_leaf[not_mask] - 1.0) / n_samples_leaf[not_mask]

        eqp2p_m0 = OnnxMax(eqp2ps_1, np.array([0], dtype=dtype), op_version=opv)
        eqp2p_m0.set_onnx_name_prefix("eqp2p_m1_%d" % i)

        eqp_ns = OnnxMul(
            OnnxDiv(
                eqp2p_m0,
                OnnxMax(eqp2ps, np.array([1], dtype=dtype), op_version=opv),
                op_version=opv,
            ),
            np.array([-2], dtype=dtype),
            op_version=opv,
        )
        eqp_ns.set_onnx_name_prefix("eqp_ns%d" % i)

        # np.ravel(node_indicator.sum(axis=1))
        # + _average_path_length(n_samples_leaf)
        # - 1.0
        av_path_length_log = OnnxMul(
            OnnxAdd(eqp_log, eqp_ns, op_version=opv), eqp2p, op_version=opv
        )
        av_path_length_log.set_onnx_name_prefix("avlog%d" % i)
        av_path_length = OnnxAdd(eq2, av_path_length_log, op_version=opv)
        av_path_length.set_onnx_name_prefix("avpl%d" % i)

        depth = OnnxAdd(
            OnnxAdd(path_length, av_path_length, op_version=opv),
            np.array([-1], dtype=dtype),
            op_version=opv,
        )
        depth.set_onnx_name_prefix("depth%d" % i)
        scores.append(depth)

    cst = len(op.estimators_) * _average_path_length([op.max_samples_])
    depths = OnnxDiv(
        OnnxSum(*scores, op_version=opv), np.array([cst], dtype=dtype), op_version=opv
    )

    # decision_function
    output_names = outputs[2].full_name if options["score_samples"] else None
    score_samples = OnnxNeg(
        OnnxPow(
            np.array([2], dtype=dtype), OnnxNeg(depths, op_version=opv), op_version=opv
        ),
        op_version=opv,
        output_names=output_names,
    )

    decision = OnnxAdd(
        score_samples,
        np.array([-op.offset_], dtype=dtype),
        op_version=opv,
        output_names=outputs[1].full_name,
    )
    decision.set_onnx_name_prefix("dec")

    less = OnnxLess(decision, np.array([0], dtype=dtype), op_version=opv)
    predict = OnnxAdd(
        OnnxMul(
            OnnxCast(less, op_version=opv, to=onnx_proto.TensorProto.INT64),
            np.array([-2], dtype=np.int64),
            op_version=opv,
        ),
        np.array([1], dtype=np.int64),
        op_version=opv,
        output_names=outputs[0].full_name,
    )
    predict.set_onnx_name_prefix("predict")

    predict.add_to(scope, container)
    less.add_to(scope, container)
    if options["score_samples"]:
        score_samples.add_to(scope, container)


def _build_labels(tree, output):
    def _recursive_build_labels(index, current):
        current[index] = True
        if tree.children_left[index] == -1:
            yield (index, current.copy())
        else:
            for it in _recursive_build_labels(tree.children_left[index], current):
                yield it
            for it in _recursive_build_labels(tree.children_right[index], current):
                yield it
        current[index] = False

    paths = {}
    current = {}

    if output == "path_length":
        for leave_index, path in _recursive_build_labels(0, current):
            spath = {}
            for nodeid, b in path.items():
                if b:
                    spath[nodeid] = 1
            paths[leave_index] = sum(spath.values())
    elif output == "node_sample":
        for leave_index, path in _recursive_build_labels(0, current):
            spath = {}
            for nodeid, b in path.items():
                if b:
                    spath[nodeid] = tree.n_node_samples[nodeid]
            paths[leave_index] = spath[leave_index]
    else:
        raise RuntimeError("Unknown method '%s'." % output)
    return paths


register_converter(
    "SklearnIsolationForest",
    convert_sklearn_isolation_forest,
    options={"score_samples": [True, False]},
)



# SPDX-License-Identifier: Apache-2.0

from ..common._apply_operation import (
    apply_concat,
    apply_abs,
    apply_add,
    apply_mul,
    apply_div,
)

try:
    from ..common._apply_operation import apply_less
except ImportError:
    # onnxconverter-common is too old
    apply_less = None
from ..common.data_types import guess_proto_type
from ..common._registration import register_converter
from ..proto import onnx_proto
from ..common._topology import Scope, Operator
from ..common._container import ModelComponentContainer


def convert_sklearn_ovr_decision_function(
    scope: Scope, operator: Operator, container: ModelComponentContainer
):
    # Applies _ovr_decision_function.
    # See https://github.com/scikit-learn/scikit-learn/blob/
    # master/sklearn/utils/multiclass.py#L407:
    # ::
    #     def _ovr_decision_function(predictions, confidences, n_classes):
    #
    #         n_samples = predictions.shape[0]
    #         votes = np.zeros((n_samples, n_classes))
    #         sum_of_confidences = np.zeros((n_samples, n_classes))
    #         k = 0
    #         for i in range(n_classes):
    #             for j in range(i + 1, n_classes):
    #                 sum_of_confidences[:, i] -= confidences[:, k]
    #                 sum_of_confidences[:, j] += confidences[:, k]
    #                 votes[predictions[:, k] == 0, i] += 1
    #                 votes[predictions[:, k] == 1, j] += 1
    #                 k += 1
    #         transformed_confidences = (
    #             sum_of_confidences / (3 * (np.abs(sum_of_confidences) + 1)))
    #         return votes + transformed_confidences
    proto_dtype = guess_proto_type(operator.inputs[1].type)
    if proto_dtype != onnx_proto.TensorProto.DOUBLE:
        proto_dtype = onnx_proto.TensorProto.FLOAT
    op = operator.raw_operator

    cst3 = scope.get_unique_variable_name("cst3")
    container.add_initializer(cst3, proto_dtype, [], [3])
    cst1 = scope.get_unique_variable_name("cst1")
    container.add_initializer(cst1, proto_dtype, [], [1])

    iprediction = operator.inputs[0].full_name
    score_name = operator.inputs[1].full_name

    n_classes = len(op.classes_)
    sumc_name = [
        scope.get_unique_variable_name("svcsumc_%d" % i) for i in range(n_classes)
    ]
    vote_name = [
        scope.get_unique_variable_name("svcvote_%d" % i) for i in range(n_classes)
    ]
    sumc_add = {n: [] for n in sumc_name}
    vote_add = {n: [] for n in vote_name}
    k = 0
    for i in range(n_classes):
        for j in range(i + 1, n_classes):
            ind = scope.get_unique_variable_name("Cind_%d" % k)
            container.add_initializer(ind, onnx_proto.TensorProto.INT64, [], [k])

            # confidences
            ext = scope.get_unique_variable_name("Csvc_%d" % k)
            container.add_node(
                "ArrayFeatureExtractor", [score_name, ind], ext, op_domain="ai.onnx.ml"
            )
            sumc_add[sumc_name[j]].append(ext)

            neg = scope.get_unique_variable_name("Cneg_%d" % k)
            container.add_node("Neg", ext, neg, op_domain="", op_version=6)
            sumc_add[sumc_name[i]].append(neg)

            # votes
            ext = scope.get_unique_variable_name("Vsvcv_%d" % k)
            container.add_node(
                "ArrayFeatureExtractor", [iprediction, ind], ext, op_domain="ai.onnx.ml"
            )
            vote_add[vote_name[j]].append(ext)

            neg = scope.get_unique_variable_name("Vnegv_%d" % k)
            container.add_node("Neg", ext, neg, op_domain="", op_version=6)
            neg1 = scope.get_unique_variable_name("Vnegv1_%d" % k)
            apply_add(
                scope,
                [neg, cst1],
                neg1,
                container,
                broadcast=1,
                operator_name="AddCl_%d_%d" % (i, j),
            )
            vote_add[vote_name[i]].append(neg1)

            # next
            k += 1

    for k, v in sumc_add.items():
        name = scope.get_unique_operator_name("Sum")
        container.add_node("Sum", v, k, op_domain="", name=name, op_version=8)
    for k, v in vote_add.items():
        name = scope.get_unique_operator_name("Sum")
        container.add_node("Sum", v, k, op_domain="", name=name, op_version=8)

    conc = scope.get_unique_variable_name("Csvcconc")
    apply_concat(scope, sumc_name, conc, container, axis=1)
    conc_vote = scope.get_unique_variable_name("Vsvcconcv")
    apply_concat(scope, vote_name, conc_vote, container, axis=1)

    conc_abs = scope.get_unique_variable_name("Cabs")
    apply_abs(scope, conc, conc_abs, container)

    conc_abs1 = scope.get_unique_variable_name("Cconc_abs1")
    apply_add(
        scope,
        [conc_abs, cst1],
        conc_abs1,
        container,
        broadcast=1,
        operator_name="AddF0",
    )
    conc_abs3 = scope.get_unique_variable_name("Cconc_abs3")
    apply_mul(scope, [conc_abs1, cst3], conc_abs3, container, broadcast=1)

    final = scope.get_unique_variable_name("Csvcfinal")
    apply_div(scope, [conc, conc_abs3], final, container, broadcast=0)

    output_name = operator.outputs[0].full_name
    apply_add(
        scope,
        [conc_vote, final],
        output_name,
        container,
        broadcast=0,
        operator_name="AddF1",
    )


register_converter("SklearnOVRDecisionFunction", convert_sklearn_ovr_decision_function)



# SPDX-License-Identifier: Apache-2.0


from ..common._apply_operation import apply_identity
from ..common._registration import register_converter
from ..common._topology import Scope, Operator
from ..common._container import ModelComponentContainer


def convert_sklearn_identity(
    scope: Scope, operator: Operator, container: ModelComponentContainer
):
    apply_identity(
        scope,
        operator.inputs[0].full_name,
        operator.outputs[0].full_name,
        container,
        operator_name=scope.get_unique_operator_name("CIdentity"),
    )


register_converter("SklearnIdentity", convert_sklearn_identity)



# SPDX-License-Identifier: Apache-2.0


import numpy as np

from ..proto import onnx_proto
from ..common._apply_operation import apply_cast, apply_concat, apply_reshape
from ..common._registration import register_converter
from ..common._topology import Scope, Operator
from ..common._container import ModelComponentContainer
from ..common.data_types import guess_proto_type, Int64TensorType
from .._supported_operators import sklearn_operator_name_map


def _fetch_scores(
    scope, container, model, inputs, raw_scores=False, is_regressor=False
):
    op_type = sklearn_operator_name_map[type(model)]
    this_operator = scope.declare_local_operator(op_type, model)
    if container.has_options(model, "raw_scores"):
        container.add_options(id(model), {"raw_scores": raw_scores})
    this_operator.inputs.append(inputs)
    if is_regressor:
        output_proba = scope.declare_local_variable("variable", inputs.type.__class__())
        this_operator.outputs.append(output_proba)
    else:
        label_name = scope.declare_local_variable("label", Int64TensorType())
        this_operator.outputs.append(label_name)
        output_proba = scope.declare_local_variable(
            "probability_tensor", inputs.type.__class__()
        )
        this_operator.outputs.append(output_proba)

    proto_type = guess_proto_type(inputs.type)
    new_name = scope.get_unique_variable_name(output_proba.full_name + "_castio")
    apply_cast(scope, output_proba.full_name, new_name, container, to=proto_type)
    return new_name


def _add_passthrough_connection(operator, predictions):
    if operator.raw_operator.passthrough:
        predictions.append(operator.inputs[0].onnx_name)


def _transform_regressor(scope, operator, container, model):
    merged_prob_tensor = scope.get_unique_variable_name("merged_probability_tensor")

    predictions = [
        _fetch_scores(scope, container, est, operator.inputs[0], is_regressor=True)
        for est in model.estimators_
    ]

    _add_passthrough_connection(operator, predictions)

    apply_concat(scope, predictions, merged_prob_tensor, container, axis=1)
    return merged_prob_tensor


def _transform(scope, operator, container, model):
    merged_prob_tensor = scope.get_unique_variable_name("merged_probability_tensor")

    predictions = [
        _fetch_scores(
            scope,
            container,
            est,
            operator.inputs[0],
            raw_scores=meth == "decision_function",
        )
        for est, meth in zip(model.estimators_, model.stack_method_)
        if est != "drop"
    ]

    op = operator.raw_operator
    select_lact_column = len(op.classes_) == 2 and all(
        op.stack_method_[est_idx] == "predict_proba"
        for est_idx in range(0, len(op.estimators_))
    )
    if select_lact_column:
        column_index_name = scope.get_unique_variable_name("column_index")
        container.add_initializer(
            column_index_name, onnx_proto.TensorProto.INT64, [], [1]
        )
        new_predictions = []
        for ipred, pred in enumerate(predictions):
            prob1 = scope.get_unique_variable_name("stack_prob%d" % ipred)
            container.add_node(
                "ArrayFeatureExtractor",
                [pred, column_index_name],
                prob1,
                name=scope.get_unique_operator_name("ArrayFeatureExtractor"),
                op_domain="ai.onnx.ml",
            )
            new_predictions.append(prob1)
        predictions = new_predictions

    _add_passthrough_connection(operator, predictions)

    apply_concat(scope, predictions, merged_prob_tensor, container, axis=1)
    return merged_prob_tensor


def convert_sklearn_stacking_classifier(
    scope: Scope, operator: Operator, container: ModelComponentContainer
):
    """
    Converter for StackingClassifier. It invokes converters for each
    estimator, concatenating their results before calling converter
    for the final estimator on the concatenated score.
    """
    stacking_op = operator.raw_operator
    classes = stacking_op.classes_
    options = container.get_options(stacking_op, dict(raw_scores=False))
    use_raw_scores = options["raw_scores"]
    class_type = onnx_proto.TensorProto.STRING
    if (
        np.issubdtype(stacking_op.classes_.dtype, np.floating)
        or stacking_op.classes_.dtype == np.bool_
    ):
        class_type = onnx_proto.TensorProto.INT32
        classes = classes.astype(np.int32)
    elif np.issubdtype(stacking_op.classes_.dtype, np.signedinteger):
        class_type = onnx_proto.TensorProto.INT32
    else:
        classes = np.array([s.encode("utf-8") for s in classes])

    classes_name = scope.get_unique_variable_name("classes")
    argmax_output_name = scope.get_unique_variable_name("argmax_output")
    reshaped_result_name = scope.get_unique_variable_name("reshaped_result")
    array_feature_extractor_result_name = scope.get_unique_variable_name(
        "array_feature_extractor_result"
    )

    container.add_initializer(classes_name, class_type, classes.shape, classes)

    merged_proba_tensor = _transform(scope, operator, container, stacking_op)
    merge_proba = scope.declare_local_variable(
        "merged_stacked_proba", operator.inputs[0].type.__class__()
    )
    container.add_node("Identity", [merged_proba_tensor], [merge_proba.onnx_name])
    prob = _fetch_scores(
        scope,
        container,
        stacking_op.final_estimator_,
        merge_proba,
        raw_scores=use_raw_scores,
    )
    container.add_node(
        "Identity",
        prob,
        operator.outputs[1].onnx_name,
        name=scope.get_unique_operator_name("OpProb"),
    )
    container.add_node(
        "ArgMax",
        prob,
        argmax_output_name,
        name=scope.get_unique_operator_name("ArgMax"),
        axis=1,
    )
    container.add_node(
        "ArrayFeatureExtractor",
        [classes_name, argmax_output_name],
        array_feature_extractor_result_name,
        op_domain="ai.onnx.ml",
        name=scope.get_unique_operator_name("ArrayFeatureExtractor"),
    )

    if class_type == onnx_proto.TensorProto.INT32:
        apply_reshape(
            scope,
            array_feature_extractor_result_name,
            reshaped_result_name,
            container,
            desired_shape=(-1,),
        )
        apply_cast(
            scope,
            reshaped_result_name,
            operator.outputs[0].full_name,
            container,
            to=onnx_proto.TensorProto.INT64,
        )
    else:
        apply_reshape(
            scope,
            array_feature_extractor_result_name,
            operator.outputs[0].full_name,
            container,
            desired_shape=(-1,),
        )


def convert_sklearn_stacking_regressor(
    scope: Scope, operator: Operator, container: ModelComponentContainer
):
    """
    Converter for StackingRegressor. It invokes converters for each
    estimator, concatenating their results before calling converter
    for the final estimator on the concatenated score.
    """
    stacking_op = operator.raw_operator

    merged_proba_tensor = _transform_regressor(scope, operator, container, stacking_op)
    merge_proba = scope.declare_local_variable(
        "merged_stacked_proba", operator.inputs[0].type.__class__()
    )
    container.add_node("Identity", [merged_proba_tensor], [merge_proba.onnx_name])
    prob = _fetch_scores(
        scope, container, stacking_op.final_estimator_, merge_proba, is_regressor=True
    )
    container.add_node(
        "Identity",
        prob,
        operator.outputs[0].full_name,
        name=scope.get_unique_operator_name("Identity"),
    )


register_converter(
    "SklearnStackingClassifier",
    convert_sklearn_stacking_classifier,
    options={
        "zipmap": [True, False, "columns"],
        "nocl": [True, False],
        "output_class_labels": [False, True],
        "raw_scores": [True, False],
    },
)
register_converter("SklearnStackingRegressor", convert_sklearn_stacking_regressor)



# SPDX-License-Identifier: Apache-2.0


from onnx.helper import make_tensor
from onnx import TensorProto
from ..proto import onnx_proto
from ..common._apply_operation import apply_concat, apply_cast
from ..common._registration import register_converter
from ..common._topology import Scope, Operator
from ..common._container import ModelComponentContainer


def convert_sklearn_polynomial_features(
    scope: Scope, operator: Operator, container: ModelComponentContainer
):
    op = operator.raw_operator
    transformed_columns = [None] * (op.n_output_features_)

    n_features = op.n_features_in_ if hasattr(op, "n_features_in_") else op.n_features_
    if hasattr(op, "_min_degree"):
        # scikit-learn >= 1.0
        combinations = op._combinations(
            n_features,
            op._min_degree,
            op._max_degree,
            op.interaction_only,
            op.include_bias,
        )
    else:
        combinations = op._combinations(
            n_features, op.degree, op.interaction_only, op.include_bias
        )

    unit_name = None
    last_feat = None
    for i, comb in enumerate(combinations):
        if len(comb) == 0:
            unit_name = scope.get_unique_variable_name("unit")
            transformed_columns[i] = unit_name
        else:
            comb_name = scope.get_unique_variable_name("comb")
            col_name = scope.get_unique_variable_name("col")
            prod_name = scope.get_unique_variable_name("prod")

            container.add_initializer(
                comb_name, onnx_proto.TensorProto.INT64, [len(comb)], list(comb)
            )

            container.add_node(
                "ArrayFeatureExtractor",
                [operator.inputs[0].full_name, comb_name],
                col_name,
                name=scope.get_unique_operator_name("ArrayFeatureExtractor"),
                op_domain="ai.onnx.ml",
            )
            reduce_prod_input = col_name
            if (
                operator.inputs[0].type._get_element_onnx_type()
                == onnx_proto.TensorProto.INT64
            ):
                float_col_name = scope.get_unique_variable_name("col")

                apply_cast(
                    scope,
                    col_name,
                    float_col_name,
                    container,
                    to=onnx_proto.TensorProto.FLOAT,
                )
                reduce_prod_input = float_col_name

            if container.target_opset >= 18:
                axis_name = scope.get_unique_variable_name("axis")
                container.add_initializer(
                    axis_name, onnx_proto.TensorProto.INT64, [1], [1]
                )
                container.add_node(
                    "ReduceProd",
                    [reduce_prod_input, axis_name],
                    prod_name,
                    name=scope.get_unique_operator_name("ReduceProd"),
                )
            else:
                container.add_node(
                    "ReduceProd",
                    reduce_prod_input,
                    prod_name,
                    axes=[1],
                    name=scope.get_unique_operator_name("ReduceProd"),
                )
            transformed_columns[i] = prod_name
            last_feat = prod_name

    if unit_name is not None:
        shape_name = scope.get_unique_variable_name("shape")
        container.add_node("Shape", last_feat, shape_name)
        container.add_node(
            "ConstantOfShape",
            shape_name,
            unit_name,
            value=make_tensor("ONE", TensorProto.FLOAT, [1], [1.0]),
            op_version=9,
        )

    if operator.inputs[0].type._get_element_onnx_type() == onnx_proto.TensorProto.INT64:
        concat_result_name = scope.get_unique_variable_name("concat_result")

        apply_concat(
            scope,
            [t for t in transformed_columns],
            concat_result_name,
            container,
            axis=1,
        )
        apply_cast(
            scope,
            concat_result_name,
            operator.outputs[0].full_name,
            container,
            to=onnx_proto.TensorProto.INT64,
        )
    else:
        apply_concat(
            scope,
            [t for t in transformed_columns],
            operator.outputs[0].full_name,
            container,
            axis=1,
        )


register_converter("SklearnPolynomialFeatures", convert_sklearn_polynomial_features)



# SPDX-License-Identifier: Apache-2.0

import numpy as np
from sklearn.preprocessing import normalize
from ..algebra.complex_functions import onnx_cdist
from ..algebra.onnx_ops import (
    OnnxMatMul,
    OnnxTranspose,
    OnnxDiv,
    OnnxSub,
    OnnxAdd,
    OnnxMul,
    OnnxPow,
    OnnxTanh,
    OnnxSqrt,
    OnnxExp,
    OnnxReduceSumApi11,
)
from ..algebra.onnx_operator import OnnxSubEstimator
from ..common._registration import register_converter
from ..common._topology import Scope, Operator
from ..common._container import ModelComponentContainer
from ..common.data_types import guess_numpy_type


def kernel_centerer_converter(
    scope: Scope, operator: Operator, container: ModelComponentContainer
):
    op = operator.raw_operator
    op_version = container.target_opset
    X = operator.inputs[0]
    dtype = guess_numpy_type(X.type)

    N = np.array([op.K_fit_rows_.shape[0]], dtype=dtype)
    K_pred_cols = OnnxDiv(
        OnnxReduceSumApi11(X, axes=[1], op_version=op_version), N, op_version=op_version
    )

    # K -= self.K_fit_rows_
    # K -= K_pred_cols
    # K += self.K_fit_all_
    K1 = OnnxSub(X, op.K_fit_rows_.astype(dtype), op_version=op_version)
    K2 = OnnxSub(K1, K_pred_cols, op_version=op_version)
    final = OnnxAdd(
        K2,
        np.array([op.K_fit_all_], dtype=dtype),
        op_version=op_version,
        output_names=operator.outputs[:1],
    )
    final.add_to(scope, container)


def kernel_pca_converter(
    scope: Scope, operator: Operator, container: ModelComponentContainer
):
    op = operator.raw_operator
    op_version = container.target_opset
    X = operator.inputs[0]
    dtype = guess_numpy_type(X.type)
    options = container.get_options(op, dict(optim=None))
    optim = options["optim"]

    # def _get_kernel(self, X, Y=None):
    # return pairwise_kernels(
    #         X, Y, metric=self.kernel, filter_params=True, **params)
    if callable(op.kernel):
        raise RuntimeError(
            "Unable to converter KernelPCA with a custom kernel %r." "" % op.kernel
        )
    if op.kernel == "precomputed":
        raise RuntimeError(
            "The converter is not implemented when kernel=%r for "
            "type=%r." % (op.kernel, type(op))
        )

    kernel = op.kernel
    params = {"gamma": op.gamma, "degree": op.degree, "coef0": op.coef0}

    if kernel == "linear":
        Y = op.X_fit_.astype(dtype)
        dist = OnnxMatMul(
            X,
            OnnxTranspose(Y, perm=[1, 0], op_version=op_version),
            op_version=op_version,
        )
    elif kernel == "cosine":
        yn = normalize(op.X_fit_, copy=True)
        ynt = yn.astype(dtype)
        norm = OnnxSqrt(
            OnnxReduceSumApi11(
                OnnxPow(X, np.array([2], dtype=np.int64), op_version=op_version),
                axes=[1],
                op_version=op_version,
                keepdims=1,
            ),
            op_version=op_version,
        )
        dist = OnnxMatMul(
            OnnxDiv(X, norm, op_version=op_version),
            OnnxTranspose(ynt, perm=[1, 0], op_version=op_version),
            op_version=op_version,
        )
    elif kernel in ("poly", "sigmoid"):
        Y = op.X_fit_.astype(dtype)
        dot = OnnxMatMul(
            X,
            OnnxTranspose(Y, perm=[1, 0], op_version=op_version),
            op_version=op_version,
        )
        if params["gamma"] is None:
            gamma = np.array([1.0 / Y.shape[1]], dtype=dtype)
        else:
            gamma = np.array([params["gamma"]], dtype=dtype)
        dot_g = OnnxMul(dot, gamma, op_version=op_version)
        dot_c = OnnxAdd(
            dot_g, np.array([params["coef0"]], dtype=dtype), op_version=op_version
        )
        if kernel == "poly":
            dist = OnnxPow(
                dot_c,
                np.array([params["degree"]], dtype=np.int64),
                op_version=op_version,
            )
        else:
            dist = OnnxTanh(dot_c, op_version=op_version)
    elif kernel == "rbf":
        if optim == "cdist":
            from skl2onnx.algebra.custom_ops import OnnxCDist

            Y = op.X_fit_.astype(dtype)
            pair = OnnxCDist(X, Y, metric="sqeuclidean", op_version=op_version)
        elif optim is None:
            Y = op.X_fit_.astype(dtype)
            dim_in = Y.shape[1] if hasattr(Y, "shape") else None
            dim_out = Y.shape[0] if hasattr(Y, "shape") else None
            pair = onnx_cdist(
                X,
                Y,
                metric="sqeuclidean",
                dtype=dtype,
                op_version=op_version,
                dim_in=dim_in,
                dim_out=dim_out,
            )
        else:
            raise ValueError("Unknown optimisation '{}'.".format(optim))
        if params["gamma"] is None:
            gamma = np.array([-1.0 / Y.shape[1]], dtype=dtype)
        else:
            gamma = np.array([-params["gamma"]], dtype=dtype)
        pair_g = OnnxMul(pair, gamma, op_version=op_version)
        dist = OnnxExp(pair_g, op_version=op_version)
    else:
        raise ValueError("Unknown kernel '{}'.".format(kernel))

    #  K = self._centerer.transform(self._get_kernel(X, self.X_fit_))
    K = OnnxSubEstimator(op._centerer, dist, op_version=op_version)

    if hasattr(op, "eigenvalues_"):
        # scikit-learn>=1.0
        non_zeros = np.flatnonzero(op.eigenvalues_)
        scaled_alphas = np.zeros_like(op.eigenvectors_)
        scaled_alphas[:, non_zeros] = op.eigenvectors_[:, non_zeros] / np.sqrt(
            op.eigenvalues_[non_zeros]
        )
    else:
        # scikit-learn<1.0
        non_zeros = np.flatnonzero(op.lambdas_)
        scaled_alphas = np.zeros_like(op.alphas_)
        scaled_alphas[:, non_zeros] = op.alphas_[:, non_zeros] / np.sqrt(
            op.lambdas_[non_zeros]
        )

    # np.dot(K, scaled_alphas)
    output = OnnxMatMul(
        K,
        scaled_alphas.astype(dtype),
        op_version=op_version,
        output_names=operator.outputs[:1],
    )

    # register the output
    output.add_to(scope, container)


register_converter("SklearnKernelCenterer", kernel_centerer_converter)
register_converter(
    "SklearnKernelPCA", kernel_pca_converter, options={"optim": [None, "cdist"]}
)



# SPDX-License-Identifier: Apache-2.0

import packaging.version as pv
import numpy as np
from onnx.helper import make_tensor
from sklearn import __version__
from ..proto import onnx_proto
from ..common._topology import Scope, Operator
from ..common._container import ModelComponentContainer
from ..common._apply_operation import (
    apply_add,
    apply_cast,
    apply_clip,
    apply_concat,
    apply_div,
    apply_exp,
    apply_mul,
    apply_reshape,
    apply_sub,
    apply_topk,
    apply_transpose,
)
from ..common.data_types import (
    FloatTensorType,
    DoubleTensorType,
    guess_proto_type,
    guess_numpy_type,
    Int64TensorType,
)
from ..common._registration import register_converter
from .._supported_operators import sklearn_operator_name_map


def _scikit_learn_before_any(any_version: str) -> bool:
    if ".dev" in __version__:
        return pv.Version(__version__.split(".dev")[0]) < pv.Version(any_version)
    if ".post" in __version__:
        return pv.Version(__version__.split(".post")[0]) < pv.Version(any_version)
    return pv.Version(__version__) < pv.Version(any_version)


def _scikit_learn_before_022() -> bool:
    return _scikit_learn_before_any("0.22")


def _scikit_learn_before_131():
    return _scikit_learn_before_any("1.3.1")


def _samme_proba(
    scope,
    container,
    proba_name,
    weight,
    zero_name_or_symmetric,
    classes_ind_name,
    one_name,
):
    weight_name = scope.get_unique_variable_name("weight")
    container.add_initializer(weight_name, onnx_proto.TensorProto.FLOAT, [], [weight])

    argmax_output_name = scope.get_unique_variable_name("argmax_output")
    container.add_node(
        "ArgMax",
        proba_name,
        argmax_output_name,
        name=scope.get_unique_operator_name("ArgMax"),
        axis=1,
    )
    equal_name = scope.get_unique_variable_name("equal")
    container.add_node(
        "Equal",
        [argmax_output_name, classes_ind_name],
        equal_name,
        name=scope.get_unique_operator_name("Equal"),
    )

    max_proba_name = scope.get_unique_variable_name("probsmax")
    container.add_node(
        "Where",
        [equal_name, one_name, zero_name_or_symmetric],
        max_proba_name,
        name=scope.get_unique_operator_name("Where"),
    )

    samme_proba_name = scope.get_unique_variable_name("samme_proba")
    apply_mul(
        scope, [max_proba_name, weight_name], samme_proba_name, container, broadcast=1
    )
    return samme_proba_name


def _samme_r_proba(scope, container, proba_name, n_classes, dtype, pdtype):
    clipped_proba_name = scope.get_unique_variable_name("clipped_proba")
    log_proba_name = scope.get_unique_variable_name("log_proba")
    reduced_proba_name = scope.get_unique_variable_name("reduced_proba")
    reshaped_result_name = scope.get_unique_variable_name("reshaped_result")
    inverted_n_classes_name = scope.get_unique_variable_name("inverted_n_classes")
    n_classes_minus_one_name = scope.get_unique_variable_name("n_classes_minus_one")
    prod_result_name = scope.get_unique_variable_name("prod_result")
    sub_result_name = scope.get_unique_variable_name("sub_result")
    samme_proba_name = scope.get_unique_variable_name("samme_proba")

    container.add_initializer(inverted_n_classes_name, pdtype, [], [1.0 / n_classes])
    container.add_initializer(n_classes_minus_one_name, pdtype, [], [n_classes - 1])

    try:
        cst_min = np.finfo(np.float64).eps.astype(dtype)
    except TypeError:
        raise TypeError(
            "Unable to convert {} (type {}) into {}.".format(
                np.finfo(float).eps, type(np.finfo(float).eps), dtype
            )
        )

    apply_clip(
        scope,
        proba_name,
        clipped_proba_name,
        container,
        operator_name=scope.get_unique_operator_name("ClipAda"),
        min=dtype(cst_min),
    )
    container.add_node(
        "Log",
        clipped_proba_name,
        log_proba_name,
        name=scope.get_unique_operator_name("Log"),
    )
    if container.target_opset < 13:
        container.add_node(
            "ReduceSum",
            log_proba_name,
            reduced_proba_name,
            axes=[1],
            name=scope.get_unique_operator_name("ReduceSum"),
        )
    else:
        axis_name = scope.get_unique_variable_name("axis")
        container.add_initializer(axis_name, onnx_proto.TensorProto.INT64, [1], [1])
        container.add_node(
            "ReduceSum",
            [log_proba_name, axis_name],
            reduced_proba_name,
            name=scope.get_unique_operator_name("ReduceSum"),
        )
    apply_reshape(
        scope,
        reduced_proba_name,
        reshaped_result_name,
        container,
        desired_shape=(-1, 1),
    )
    apply_mul(
        scope,
        [reshaped_result_name, inverted_n_classes_name],
        prod_result_name,
        container,
        broadcast=1,
    )
    apply_sub(
        scope,
        [log_proba_name, prod_result_name],
        sub_result_name,
        container,
        broadcast=1,
    )
    apply_mul(
        scope,
        [sub_result_name, n_classes_minus_one_name],
        samme_proba_name,
        container,
        broadcast=1,
    )
    return samme_proba_name


def _normalise_probability(scope, container, operator, proba_names_list, model):
    est_weights_sum_name = scope.get_unique_variable_name("est_weights_sum")
    summation_prob_name = scope.get_unique_variable_name("summation_prob")
    div_result_name = scope.get_unique_variable_name("div_result")
    exp_operand_name = scope.get_unique_variable_name("exp_operand")
    exp_result_name = scope.get_unique_variable_name("exp_result")
    reduced_exp_result_name = scope.get_unique_variable_name("reduced_exp_result")
    normaliser_name = scope.get_unique_variable_name("normaliser")
    zero_scalar_name = scope.get_unique_variable_name("zero_scalar")
    comparison_result_name = scope.get_unique_variable_name("comparison_result")
    cast_output_name = scope.get_unique_variable_name("cast_output")
    zero_filtered_normaliser_name = scope.get_unique_variable_name(
        "zero_filtered_normaliser"
    )
    mul_operand_name = scope.get_unique_variable_name("mul_operand")
    cast_normaliser_name = scope.get_unique_variable_name("cast_normaliser")

    proto_dtype = guess_proto_type(operator.inputs[0].type)
    if proto_dtype != onnx_proto.TensorProto.DOUBLE:
        proto_dtype = onnx_proto.TensorProto.FLOAT

    container.add_initializer(
        est_weights_sum_name, proto_dtype, [], [model.estimator_weights_.sum()]
    )
    container.add_initializer(
        mul_operand_name, proto_dtype, [], [1.0 / (model.n_classes_ - 1)]
    )
    container.add_initializer(zero_scalar_name, onnx_proto.TensorProto.INT32, [], [0])

    container.add_node(
        "Sum",
        proba_names_list,
        summation_prob_name,
        name=scope.get_unique_operator_name("Sum"),
    )
    apply_div(
        scope,
        [summation_prob_name, est_weights_sum_name],
        div_result_name,
        container,
        broadcast=1,
    )
    apply_mul(
        scope,
        [div_result_name, mul_operand_name],
        exp_operand_name,
        container,
        broadcast=1,
    )
    apply_exp(scope, exp_operand_name, exp_result_name, container)
    if container.target_opset < 13:
        container.add_node(
            "ReduceSum",
            exp_result_name,
            reduced_exp_result_name,
            axes=[1],
            name=scope.get_unique_operator_name("ReduceSum"),
        )
    else:
        axis_name = scope.get_unique_variable_name("axis")
        container.add_initializer(axis_name, onnx_proto.TensorProto.INT64, [1], [1])
        container.add_node(
            "ReduceSum",
            [exp_result_name, axis_name],
            reduced_exp_result_name,
            name=scope.get_unique_operator_name("ReduceSum"),
        )
    apply_reshape(
        scope,
        reduced_exp_result_name,
        normaliser_name,
        container,
        desired_shape=(-1, 1),
    )
    apply_cast(
        scope,
        normaliser_name,
        cast_normaliser_name,
        container,
        to=onnx_proto.TensorProto.INT32,
    )
    container.add_node(
        "Equal",
        [cast_normaliser_name, zero_scalar_name],
        comparison_result_name,
        name=scope.get_unique_operator_name("Equal"),
    )
    apply_cast(
        scope, comparison_result_name, cast_output_name, container, to=proto_dtype
    )
    apply_add(
        scope,
        [normaliser_name, cast_output_name],
        zero_filtered_normaliser_name,
        container,
        broadcast=0,
    )
    apply_div(
        scope,
        [exp_result_name, zero_filtered_normaliser_name],
        operator.outputs[1].full_name,
        container,
        broadcast=1,
    )
    return operator.outputs[1].full_name


def _generate_raw_scores(scope, container, operator, proba_names_list, model):
    summation_prob_name = scope.get_unique_variable_name("summation_proba")
    est_weights_sum_name = scope.get_unique_variable_name("est_weights")

    proto_dtype = guess_proto_type(operator.inputs[0].type)
    if proto_dtype != onnx_proto.TensorProto.DOUBLE:
        proto_dtype = onnx_proto.TensorProto.FLOAT

    container.add_initializer(
        est_weights_sum_name, proto_dtype, [], [model.estimator_weights_.sum()]
    )

    container.add_node(
        "Sum",
        proba_names_list,
        summation_prob_name,
        name=scope.get_unique_operator_name("Sum"),
    )

    if len(model.classes_) == 2:
        div_res_name = scope.get_unique_variable_name("div_res")
        operand_name = scope.get_unique_variable_name("operand")
        neg_name = scope.get_unique_variable_name("neg")
        mul_res_name = scope.get_unique_variable_name("mul_res")
        pos_class_scores_name = scope.get_unique_variable_name("pos_class_scores")
        neg_class_scores_name = scope.get_unique_variable_name("neg_class_scores")
        container.add_initializer(operand_name, proto_dtype, [2], [-1, 1])
        container.add_initializer(neg_name, proto_dtype, [], [-1])

        apply_div(
            scope,
            [summation_prob_name, est_weights_sum_name],
            div_res_name,
            container,
            broadcast=1,
        )
        apply_mul(
            scope, [div_res_name, operand_name], mul_res_name, container, broadcast=1
        )
        if container.target_opset < 13:
            container.add_node(
                "ReduceSum",
                mul_res_name,
                pos_class_scores_name,
                axes=[1],
                name=scope.get_unique_operator_name("ReduceSum"),
            )
        else:
            axis_name = scope.get_unique_variable_name("axis")
            container.add_initializer(axis_name, onnx_proto.TensorProto.INT64, [1], [1])
            container.add_node(
                "ReduceSum",
                [mul_res_name, axis_name],
                pos_class_scores_name,
                name=scope.get_unique_operator_name("ReduceSum"),
            )
        apply_mul(
            scope,
            [pos_class_scores_name, neg_name],
            neg_class_scores_name,
            container,
            broadcast=1,
        )
        apply_concat(
            scope,
            [neg_class_scores_name, pos_class_scores_name],
            operator.outputs[1].full_name,
            container,
            axis=1,
        )
    else:
        apply_div(
            scope,
            [summation_prob_name, est_weights_sum_name],
            operator.outputs[1].full_name,
            container,
            broadcast=1,
        )
    return operator.outputs[1].full_name


def convert_sklearn_ada_boost_classifier(
    scope: Scope, operator: Operator, container: ModelComponentContainer
):
    """
    Converter for AdaBoost classifier.
    This function goes through the list of estimators and uses
    TreeEnsembleClassifer op to calculate class probabilities
    for each estimator. Then it calculates the weighted sum
    across all the estimators depending on the algorithm
    picked during trainging (SAMME.R or SAMME) and normalises
    the probability score for the final result. Label is
    calculated by simply doing an argmax of the probability scores.
    """
    if scope.get_options(operator.raw_operator, dict(nocl=False))["nocl"]:
        raise RuntimeError(
            "Option 'nocl' is not implemented for operator '{}'.".format(
                operator.raw_operator.__class__.__name__
            )
        )
    op = operator.raw_operator
    options = container.get_options(op, dict(raw_scores=False))
    use_raw_scores = options["raw_scores"]
    classes = op.classes_
    class_type = onnx_proto.TensorProto.STRING
    if np.issubdtype(classes.dtype, np.floating):
        class_type = onnx_proto.TensorProto.INT32
        classes = classes.astype("int")
    elif np.issubdtype(classes.dtype, np.signedinteger):
        class_type = onnx_proto.TensorProto.INT32
    else:
        classes = np.array([s.encode("utf-8") for s in classes])

    argmax_output_name = scope.get_unique_variable_name("argmax_output")
    array_feature_extractor_result_name = scope.get_unique_variable_name(
        "array_feature_extractor_result"
    )

    classes_name = scope.get_unique_variable_name("classes")
    container.add_initializer(classes_name, class_type, classes.shape, classes)

    proba_names_list = []
    classes_ind_name = None
    zero_name = None
    one_name = None
    classes_ind_name = None
    symmetric_weight = None

    proto_dtype = guess_proto_type(operator.inputs[0].type)
    proba_type = operator.inputs[0].type.__class__
    add_cast = proto_dtype == onnx_proto.TensorProto.INT64
    if proto_dtype != onnx_proto.TensorProto.DOUBLE:
        proto_dtype = onnx_proto.TensorProto.FLOAT
    dtype = guess_numpy_type(operator.inputs[0].type)
    if dtype != np.float64:
        dtype = np.float32

    for i_est, estimator in enumerate(op.estimators_):
        label_name = scope.declare_local_variable(
            "elab_name_%d" % i_est, Int64TensorType()
        )
        proba_name = scope.declare_local_variable("eprob_name_%d" % i_est, proba_type())

        op_type = sklearn_operator_name_map[type(estimator)]

        this_operator = scope.declare_local_operator(op_type, estimator)
        this_operator.inputs = operator.inputs
        this_operator.outputs.extend([label_name, proba_name])

        if add_cast:
            this_operator = scope.declare_local_operator("SklearnCast")
            this_operator.inputs.append(proba_name)
            var_name = scope.declare_local_variable("cast", FloatTensorType())
            this_operator.outputs.append(var_name)
            proba_name = var_name

        if op.algorithm == "SAMME.R":
            cur_proba_name = _samme_r_proba(
                scope, container, proba_name.onnx_name, len(classes), dtype, proto_dtype
            )
        else:
            # SAMME
            if _scikit_learn_before_022() and not use_raw_scores:
                weight_name = scope.get_unique_variable_name("weight")
                samme_proba_name = scope.get_unique_variable_name("samme_proba")
                container.add_initializer(
                    weight_name,
                    onnx_proto.TensorProto.FLOAT,
                    [],
                    [op.estimator_weights_[i_est]],
                )
                apply_mul(
                    scope,
                    [proba_name.onnx_name, weight_name],
                    samme_proba_name,
                    container,
                    broadcast=1,
                )
                cur_proba_name = samme_proba_name
            else:
                if classes_ind_name is None:
                    classes_ind_name = scope.get_unique_variable_name("classes_ind3")
                    container.add_initializer(
                        classes_ind_name,
                        onnx_proto.TensorProto.INT64,
                        (1, len(classes)),
                        list(range(len(classes))),
                    )

                if zero_name is None:
                    shape_name = scope.get_unique_variable_name("shape")
                    container.add_node(
                        "Shape",
                        proba_name.onnx_name,
                        shape_name,
                        name=scope.get_unique_operator_name("Shape"),
                    )

                    zero_name = scope.get_unique_variable_name("zero")
                    container.add_node(
                        "ConstantOfShape",
                        shape_name,
                        zero_name,
                        name=scope.get_unique_operator_name("CoSA"),
                        value=make_tensor(
                            "value", onnx_proto.TensorProto.FLOAT, (1,), [0]
                        ),
                    )

                    one_name = scope.get_unique_variable_name("one")
                    container.add_node(
                        "ConstantOfShape",
                        shape_name,
                        one_name,
                        name=scope.get_unique_operator_name("CoSB"),
                        value=make_tensor(
                            "value", onnx_proto.TensorProto.FLOAT, (1,), [1.0]
                        ),
                    )

                if not _scikit_learn_before_131() and symmetric_weight is None:
                    symmetric_weight = scope.get_unique_variable_name(
                        "symmetrix_weight"
                    )
                    container.add_initializer(
                        symmetric_weight, proto_dtype, [1], [-1 / (len(classes) - 1)]
                    )

                cur_proba_name = _samme_proba(
                    scope,
                    container,
                    proba_name.onnx_name,
                    op.estimator_weights_[i_est],
                    # See https://github.com/scikit-learn/scikit-learn/pull/26521
                    zero_name if _scikit_learn_before_131() else symmetric_weight,
                    classes_ind_name,
                    one_name,
                )

        proba_names_list.append(cur_proba_name)

    function = _generate_raw_scores if use_raw_scores else _normalise_probability
    class_prob_name = function(scope, container, operator, proba_names_list, op)
    container.add_node(
        "ArgMax",
        class_prob_name,
        argmax_output_name,
        name=scope.get_unique_operator_name("ArgMax"),
        axis=1,
    )
    container.add_node(
        "ArrayFeatureExtractor",
        [classes_name, argmax_output_name],
        array_feature_extractor_result_name,
        op_domain="ai.onnx.ml",
        name=scope.get_unique_operator_name("ArrayFeatureExtractor"),
    )

    if class_type == onnx_proto.TensorProto.INT32:
        reshaped_result_name = scope.get_unique_variable_name("reshaped_result")

        apply_reshape(
            scope,
            array_feature_extractor_result_name,
            reshaped_result_name,
            container,
            desired_shape=(-1,),
        )
        apply_cast(
            scope,
            reshaped_result_name,
            operator.outputs[0].full_name,
            container,
            to=onnx_proto.TensorProto.INT64,
        )
    else:
        apply_reshape(
            scope,
            array_feature_extractor_result_name,
            operator.outputs[0].full_name,
            container,
            desired_shape=(-1,),
        )


def _get_estimators_label(scope, operator, container, model):
    """
    This function computes labels for each estimator and returns
    a tensor produced by concatenating the labels.
    """
    if isinstance(operator.inputs[0].type, DoubleTensorType):
        var_type = DoubleTensorType
    else:
        var_type = FloatTensorType
    concatenated_labels_name = scope.get_unique_variable_name("concatenated_labels")

    input_name = operator.inputs
    estimators_results_list = []
    for i, estimator in enumerate(model.estimators_):
        estimator_label_name = scope.declare_local_variable(
            "est_label_%d" % i, var_type([None, 1])
        )

        op_type = sklearn_operator_name_map[type(estimator)]

        this_operator = scope.declare_local_operator(op_type, estimator)
        this_operator.inputs = input_name
        this_operator.outputs.append(estimator_label_name)

        estimators_results_list.append(estimator_label_name.onnx_name)

    apply_concat(
        scope, estimators_results_list, concatenated_labels_name, container, axis=1
    )
    return concatenated_labels_name


def cum_sum(scope, container, rnn_input_name, sequence_length, proto_dtype):
    opv = container.target_opset
    weights_cdf_name = scope.get_unique_variable_name("weights_cdf")
    if opv < 11:
        transposed_input_name = scope.get_unique_variable_name("transposed_input")
        reshaped_result_name = scope.get_unique_variable_name("reshaped_result")
        weights_name = scope.get_unique_variable_name("weights")
        rec_weights_name = scope.get_unique_variable_name("rec_weights")
        rnn_output_name = scope.get_unique_variable_name("rnn_output")
        permuted_rnn_y_name = scope.get_unique_variable_name("permuted_rnn_y")

        container.add_initializer(weights_name, proto_dtype, [1, 1, 1], [1])
        container.add_initializer(rec_weights_name, proto_dtype, [1, 1, 1], [1])

        apply_transpose(
            scope, rnn_input_name, transposed_input_name, container, perm=(1, 0)
        )
        apply_reshape(
            scope,
            transposed_input_name,
            reshaped_result_name,
            container,
            desired_shape=(sequence_length, -1, 1),
        )
        container.add_node(
            "RNN",
            inputs=[reshaped_result_name, weights_name, rec_weights_name],
            outputs=[rnn_output_name],
            activations=["Affine"],
            name=scope.get_unique_operator_name("RNN"),
            activation_alpha=[1.0],
            activation_beta=[0.0],
            hidden_size=1,
        )
        apply_transpose(
            scope, rnn_output_name, permuted_rnn_y_name, container, perm=(2, 0, 1, 3)
        )
        apply_reshape(
            scope,
            permuted_rnn_y_name,
            weights_cdf_name,
            container,
            desired_shape=(-1, sequence_length),
        )
    else:
        axis_name = scope.get_unique_variable_name("axis_name")
        container.add_initializer(axis_name, onnx_proto.TensorProto.INT32, [], [1])
        container.add_node(
            "CumSum",
            [rnn_input_name, axis_name],
            [weights_cdf_name],
            name=scope.get_unique_operator_name("CumSum"),
            op_version=11,
        )
    return weights_cdf_name


def _apply_gather_elements(
    scope, container, inputs, output, axis, dim, zero_type, suffix
):
    if container.target_opset >= 11:
        container.add_node(
            "GatherElements",
            inputs,
            output,
            op_version=11,
            axis=axis,
            name=scope.get_unique_operator_name("GatEls" + suffix),
        )
    else:
        classes_ind_name = scope.get_unique_variable_name("classes_ind2")
        container.add_initializer(
            classes_ind_name, onnx_proto.TensorProto.INT64, (1, dim), list(range(dim))
        )

        shape_name = scope.get_unique_variable_name("shape")
        container.add_node(
            "Shape", inputs[0], shape_name, name=scope.get_unique_operator_name("Shape")
        )
        zero_name = scope.get_unique_variable_name("zero")
        zero_val = 0 if zero_type == onnx_proto.TensorProto.INT64 else 0.0
        container.add_node(
            "ConstantOfShape",
            shape_name,
            zero_name,
            name=scope.get_unique_operator_name("CoSA"),
            value=make_tensor("value", zero_type, (1,), [zero_val]),
            op_version=9,
        )

        equal_name = scope.get_unique_variable_name("equal")
        container.add_node(
            "Equal",
            [inputs[1], classes_ind_name],
            equal_name,
            name=scope.get_unique_operator_name("Equal"),
        )

        selected = scope.get_unique_variable_name("selected")
        container.add_node(
            "Where",
            [equal_name, inputs[0], zero_name],
            selected,
            name=scope.get_unique_operator_name("Where"),
        )
        if container.target_opset < 13:
            container.add_node(
                "ReduceSum",
                selected,
                output,
                axes=[1],
                name=scope.get_unique_operator_name("ReduceSum"),
            )
        else:
            axis_name = scope.get_unique_variable_name("axis")
            container.add_initializer(axis_name, onnx_proto.TensorProto.INT64, [1], [1])
            container.add_node(
                "ReduceSum",
                [selected, axis_name],
                output,
                name=scope.get_unique_operator_name("ReduceSum"),
            )


def convert_sklearn_ada_boost_regressor(
    scope: Scope, operator: Operator, container: ModelComponentContainer
):
    """
    Converter for AdaBoost regressor.
    This function first calls _get_estimators_label() which returns a
    tensor of concatenated labels predicted by each estimator. Then,
    median is calculated and returned as the final output.

    Note: This function creates an ONNX model which can predict on only
    one instance at a time because ArrayFeatureExtractor can only
    extract based on the last axis, so we can't fetch different columns
    for different rows.
    """
    proto_dtype = guess_proto_type(operator.inputs[0].type)
    if proto_dtype != onnx_proto.TensorProto.DOUBLE:
        proto_dtype = onnx_proto.TensorProto.FLOAT

    op = operator.raw_operator

    negate_name = scope.get_unique_variable_name("negate")
    estimators_weights_name = scope.get_unique_variable_name("estimators_weights")
    half_scalar_name = scope.get_unique_variable_name("half_scalar")
    last_index_name = scope.get_unique_variable_name("last_index")
    negated_labels_name = scope.get_unique_variable_name("negated_labels")
    sorted_values_name = scope.get_unique_variable_name("sorted_values")
    sorted_indices_name = scope.get_unique_variable_name("sorted_indices")
    array_feat_extractor_output_name = scope.get_unique_variable_name(
        "array_feat_extractor_output"
    )
    median_value_name = scope.get_unique_variable_name("median_value")
    comp_value_name = scope.get_unique_variable_name("comp_value")
    median_or_above_name = scope.get_unique_variable_name("median_or_above")
    median_idx_name = scope.get_unique_variable_name("median_idx")
    cast_result_name = scope.get_unique_variable_name("cast_result")
    reshaped_weights_name = scope.get_unique_variable_name("reshaped_weights")
    median_estimators_name = scope.get_unique_variable_name("median_estimators")

    container.add_initializer(negate_name, proto_dtype, [], [-1])
    container.add_initializer(
        estimators_weights_name,
        proto_dtype,
        [len(op.estimator_weights_)],
        op.estimator_weights_,
    )
    container.add_initializer(half_scalar_name, proto_dtype, [], [0.5])
    container.add_initializer(
        last_index_name, onnx_proto.TensorProto.INT64, [], [len(op.estimators_) - 1]
    )

    concatenated_labels = _get_estimators_label(scope, operator, container, op)
    apply_mul(
        scope,
        [concatenated_labels, negate_name],
        negated_labels_name,
        container,
        broadcast=1,
    )
    try:
        apply_topk(
            scope,
            negated_labels_name,
            [sorted_values_name, sorted_indices_name],
            container,
            k=len(op.estimators_),
        )
    except TypeError:
        # onnxconverter-common < 1.7.0
        apply_topk(
            scope,
            [negated_labels_name],
            [sorted_values_name, sorted_indices_name],
            container,
            k=len(op.estimators_),
        )
    container.add_node(
        "ArrayFeatureExtractor",
        [estimators_weights_name, sorted_indices_name],
        array_feat_extractor_output_name,
        op_domain="ai.onnx.ml",
        name=scope.get_unique_operator_name("ArrayFeatureExtractor"),
    )
    apply_reshape(
        scope,
        array_feat_extractor_output_name,
        reshaped_weights_name,
        container,
        desired_shape=(-1, len(op.estimators_)),
    )
    weights_cdf_name = cum_sum(
        scope, container, reshaped_weights_name, len(op.estimators_), proto_dtype
    )
    container.add_node(
        "ArrayFeatureExtractor",
        [weights_cdf_name, last_index_name],
        median_value_name,
        op_domain="ai.onnx.ml",
        name=scope.get_unique_operator_name("ArrayFeatureExtractor"),
    )
    apply_mul(
        scope,
        [median_value_name, half_scalar_name],
        comp_value_name,
        container,
        broadcast=1,
    )
    container.add_node(
        "Less",
        [weights_cdf_name, comp_value_name],
        median_or_above_name,
        name=scope.get_unique_operator_name("Less"),
    )
    apply_cast(scope, median_or_above_name, cast_result_name, container, to=proto_dtype)
    container.add_node(
        "ArgMin",
        cast_result_name,
        median_idx_name,
        name=scope.get_unique_operator_name("ArgMin"),
        axis=1,
    )
    _apply_gather_elements(
        scope,
        container,
        [sorted_indices_name, median_idx_name],
        median_estimators_name,
        axis=1,
        dim=len(op.estimators_),
        zero_type=onnx_proto.TensorProto.INT64,
        suffix="A",
    )
    output_name = operator.output_full_names[0]
    _apply_gather_elements(
        scope,
        container,
        [concatenated_labels, median_estimators_name],
        output_name,
        axis=1,
        dim=len(op.estimators_),
        zero_type=proto_dtype,
        suffix="B",
    )


register_converter(
    "SklearnAdaBoostClassifier",
    convert_sklearn_ada_boost_classifier,
    options={
        "zipmap": [True, False, "columns"],
        "nocl": [True, False],
        "output_class_labels": [False, True],
        "raw_scores": [True, False],
    },
)
register_converter("SklearnAdaBoostRegressor", convert_sklearn_ada_boost_regressor)



# SPDX-License-Identifier: Apache-2.0

import numpy as np
from ..common._registration import register_converter
from ..common._topology import Scope, Operator
from ..common._container import ModelComponentContainer
from ..algebra.onnx_operator import OnnxSubEstimator
from ..algebra.onnx_ops import OnnxIdentity, OnnxConcat, OnnxReshape


def convert_sklearn_random_tree_embedding(
    scope: Scope, operator: Operator, container: ModelComponentContainer
):
    X = operator.inputs[0]
    out = operator.outputs
    op = operator.raw_operator
    opv = container.target_opset

    if op.sparse_output:
        raise RuntimeError(
            "The converter cannot convert the model with sparse outputs."
        )

    outputs = []
    for est in op.estimators_:
        leave = OnnxSubEstimator(
            est, X, op_version=opv, options={"decision_leaf": True}
        )
        outputs.append(
            OnnxReshape(leave[1], np.array([-1, 1], dtype=np.int64), op_version=opv)
        )
    merged = OnnxConcat(*outputs, axis=1, op_version=opv)
    ohe = OnnxSubEstimator(op.one_hot_encoder_, merged, op_version=opv)
    y = OnnxIdentity(ohe, op_version=opv, output_names=out)
    y.add_to(scope, container)


register_converter("SklearnRandomTreesEmbedding", convert_sklearn_random_tree_embedding)
